from ..spec import AbstractFileSystem
from ..utils import infer_storage_options
from pyarrow.hdfs import HadoopFileSystem


class PyArrowHDFS(AbstractFileSystem):
    """Adapted version of Arrow's HadoopFileSystem

    This is a very simple wrapper over pa.hdfs.HadoopFileSystem, which
    passes on all calls to the underlying class.
    """

    def __init__(self, host="default", port=0, user=None, kerb_ticket=None,
                 driver='libhdfs', extra_conf=None, **kwargs):
        """

        Parameters
        ----------
        host: str
            Hostname, IP or "default" to try to read from Hadoop config
        port: int
            Port to connect on, or default from Hadoop config if 0
        user: str or None
            If given, connect as this username
        kerb_ticket: str or None
            If given, use this ticket for authentication
        driver: 'libhdfs' or 'libhdfs3'
            Binary driver; libhdfs if the JNI library and default
        extra_conf: None or dict
            Passed on to HadoopFileSystem
        """
        AbstractFileSystem.__init__(self, **kwargs)
        self.pars = (host, port, user, kerb_ticket, driver, extra_conf)
        self.pahdfs = HadoopFileSystem(host=host, port=port, user=user,
                                       kerb_ticket=kerb_ticket, driver=driver,
                                       extra_conf=extra_conf)

    def _open(self, path, mode='rb', block_size=None, autocommit=True,
              **kwargs):
        """

        Parameters
        ----------
        path: str
            Location of file; should start with '/'
        mode: str
        block_size: int
            Hadoop block size, e.g., 2**26
        autocommit: True
            Transactions are not yet implemented for HDFS; errors if not True
        kwargs: dict or None
            Hadoop config parameters

        Returns
        -------
        HDFSFile file-like instance
        """
        if not autocommit:
            raise NotImplementedError
        return HDFSFile(self, path, mode, block_size, **kwargs)

    def __reduce_ex__(self, protocol):
        return PyArrowHDFS, self.pars

    def ls(self, path, detail=True):
        out = self.pahdfs.ls(path, detail)
        if detail:
            for p in out:
                p['type'] = p['kind']
                p["name"] = self._strip_protocol(p["name"])
        else:
            out = [self._strip_protocol(p) for p in out]
        return out

    @staticmethod
    def _get_kwargs_from_urls(paths):
        ops = infer_storage_options(paths)
        out = {}
        if ops.get('host', None):
            out['host'] = ops['host']
        if ops.get('username', None):
            out['user'] = ops['username']
        if ops.get('port', None):
            out['port'] = ops['port']
        return out

    @classmethod
    def _strip_protocol(cls, path):
        ops = infer_storage_options(path)
        return ops['path']

    def __getattribute__(self, item):
        if item in ['_open', '__init__', '__getattribute__', '__reduce_ex__',
                    'open', 'ls', 'makedirs']:
            # all the methods defined in this class. Note `open` here, since
            # it calls `_open`, but is actually in superclass
            return lambda *args, **kw: getattr(PyArrowHDFS, item)(
                self, *args, **kw
            )
        if item == '__class__':
            return PyArrowHDFS
        d = object.__getattribute__(self, '__dict__')
        pahdfs = d.get('pahdfs', None)  # fs is not immediately defined
        if pahdfs is not None and item in [
            'chmod', 'chown', 'user',
            'df', 'disk_usage', 'download', 'driver', 'exists',
            'extra_conf', 'get_capacity', 'get_space_used', 'host',
            'is_open', 'kerb_ticket', 'strip_protocol',
            'mkdir', 'mv', 'port', 'get_capacity',
            'get_space_used', 'df', 'chmod', 'chown', 'disk_usage',
            'download', 'upload', '_get_kwargs_from_urls',
            'read_parquet', 'rm', 'stat', 'upload',
        ]:
            return getattr(pahdfs, item)
        else:
            # attributes of the superclass, while target is being set up
            return super().__getattribute__(item)


class HDFSFile(object):
    """Wrapper around arrow's HdfsFile

    Allows seek beyond EOF and (eventually) commit/discard
    """

    def __init__(self, fs, path, mode, block_size, **kwargs):
        self.fs = fs
        self.path = path
        self.mode = mode
        self.block_size = block_size
        self.fh = fs.pahdfs.open(path, mode, block_size, **kwargs)
        if self.fh.readable():
            self.seek_size = self.size()

    def seek(self, loc, whence=0):
        if whence == 0 and self.readable():
            loc = min(loc, self.seek_size)
        return self.fh.seek(loc, whence)

    def __getattr__(self, item):
        return getattr(self.fh, item)

    def __reduce_ex__(self, protocol):
        return HDFSFile, (self.fs, self.path, self.mode, self.block_size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""Tests abstract buffered file API, using FTP implementation"""
import pickle
import pytest
from fsspec.implementations.tests.test_ftp import ftp_writable, FTPFileSystem
data = b'hello' * 10000


def test_pickle(ftp_writable):
    host, port, user, pw = ftp_writable
    ftp = FTPFileSystem(host=host, port=port,
                        username=user, password=pw)

    f = ftp.open('/out', 'rb')

    f2 = pickle.loads(pickle.dumps(f))
    assert f == f2


def test_file_read_attributes(ftp_writable):
    host, port, user, pw = ftp_writable
    ftp = FTPFileSystem(host=host, port=port,
                        username=user, password=pw)

    f = ftp.open('/out', 'rb')
    assert f.info()['size'] == len(data)
    assert f.tell() == 0
    assert f.seekable()
    assert f.readable()
    assert not f.writable()
    out = bytearray(len(data))

    assert f.read() == data
    assert f.read() == b""
    f.seek(0)
    assert f.readuntil(b'l') == b'hel'
    assert f.tell() == 3

    f.readinto1(out)
    assert out[:-3] == data[3:]
    with pytest.raises(ValueError):
        f.write(b'')
    f.close()
    with pytest.raises(ValueError):
        f.read()(b'')


def test_seek(ftp_writable):
    host, port, user, pw = ftp_writable
    ftp = FTPFileSystem(host=host, port=port,
                        username=user, password=pw)

    f = ftp.open('/out', 'rb')

    assert f.seek(-10, 2) == len(data) - 10
    assert f.tell() == len(data) - 10
    assert f.seek(-1, 1) == len(data) - 11
    with pytest.raises(ValueError):
        f.seek(-1)
    with pytest.raises(ValueError):
        f.seek(0, 7)


def test_file_idempotent(ftp_writable):
    host, port, user, pw = ftp_writable
    ftp = FTPFileSystem(host=host, port=port,
                        username=user, password=pw)

    f = ftp.open('/out', 'rb')
    f2 = ftp.open('/out', 'rb')
    assert hash(f) == hash(f2)
    assert f == f2
    ftp.touch('/out2')
    f2 = ftp.open('/out2', 'rb')
    assert hash(f2) != hash(f)
    assert f != f2
    f2 = ftp.open('/out', 'wb')
    assert hash(f2) != hash(f)


def test_file_text_attributes(ftp_writable):
    host, port, user, pw = ftp_writable
    ftp = FTPFileSystem(host=host, port=port,
                        username=user, password=pw)

    data = b'hello\n' * 1000
    with ftp.open('/out2', 'wb') as f:
        f.write(data)

    f = ftp.open('/out2', 'rb')
    assert f.readline() == b'hello\n'
    f.seek(0)
    assert list(f) == [d + b'\n' for d in data.split()]
    f.seek(0)
    assert f.readlines() == [d + b'\n' for d in data.split()]

    f = ftp.open('/out2', 'rt')
    assert f.readline() == 'hello\n'
    assert f.encoding


def test_file_write_attributes(ftp_writable):
    host, port, user, pw = ftp_writable
    ftp = FTPFileSystem(host=host, port=port,
                        username=user, password=pw)
    f = ftp.open('/out2', 'wb')
    with pytest.raises(ValueError):
        f.info()
    with pytest.raises(ValueError):
        f.seek(0)
    with pytest.raises(ValueError):
        f.read(0)
    assert not f.readable()
    assert f.writable()

    f.flush()  # no-op

    assert f.write(b'hello') == 5
    assert f.write(b'hello') == 5
    assert not f.closed
    f.close()
    assert f.closed
    with pytest.raises(ValueError):
        f.write(b'')
    with pytest.raises(ValueError):
        f.flush()


def test_midread_cache(ftp_writable):
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host=host, port=port,
                       username=user, password=pw)
    fn = "/myfile"
    with fs.open(fn, 'wb') as f:
        f.write(b'a' * 175_627_146)
    with fs.open(fn, 'rb') as f:
        f.seek(175561610)
        d1 = f.read(65536)
        assert len(d1) == 65536

        f.seek(4)
        size = 17562198
        d2 = f.read(size)
        assert len(d2) == size

        f.seek(17562288)
        size = 17562187
        d3 = f.read(size)
        assert len(d3) == size


def test_read_block(ftp_writable):
    # not the same as test_read_block in test_utils, this depends on the
    # behaviour of the bytest caching
    from fsspec.utils import read_block
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host=host, port=port,
                       username=user, password=pw)
    fn = "/myfile"
    with fs.open(fn, 'wb') as f:
        f.write(b'a,b\n1,2')
    f = fs.open(fn, 'rb', cache_type='bytes')
    assert read_block(f, 0, 6400, b'\n') == b'a,b\n1,2'


def test_with_gzip(ftp_writable):
    import gzip
    data = b'some compressable stuff'
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host=host, port=port,
                       username=user, password=pw)
    fn = "/myfile"
    with fs.open(fn, 'wb') as f:
        gf = gzip.GzipFile(fileobj=f, mode='w')
        gf.write(data)
        gf.close()
    with fs.open(fn, 'rb') as f:
        gf = gzip.GzipFile(fileobj=f, mode='r')
        assert gf.read() == data



import os
import pytest
import shutil
import subprocess
import sys
import time
from fsspec.implementations.ftp import FTPFileSystem
from fsspec import open_files
import fsspec

pytest.importorskip('pyftpdlib')
here = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture()
def ftp():
    P = subprocess.Popen([sys.executable, '-m', 'pyftpdlib', '-d', here],
                         stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    try:
        time.sleep(1)
        yield 'localhost', 2121
    finally:
        P.terminate()
        P.wait()


@pytest.fixture()
def ftp_writable(tmpdir):
    d = str(tmpdir)
    with open(os.path.join(d, 'out'), 'wb') as f:
        f.write(b'hello' * 10000)
    P = subprocess.Popen([sys.executable, '-m', 'pyftpdlib', '-d', d,
                          '-u', 'user', '-P', 'pass', '-w'])
    try:
        time.sleep(1)
        yield 'localhost', 2121, 'user', 'pass'
    finally:
        P.terminate()
        P.wait()
        try:
            shutil.rmtree(tmpdir)
        except:
            pass


def test_basic(ftp):
    host, port = ftp
    fs = FTPFileSystem(host, port)
    assert fs.ls('/', detail=False) == sorted(os.listdir(here))
    out = fs.cat('/' + os.path.basename(__file__))
    assert out == open(__file__, 'rb').read()


@pytest.mark.parametrize('cache_type', ['bytes', 'mmap'])
def test_complex(ftp_writable, cache_type):
    from fsspec.core import BytesCache
    host, port, user, pw = ftp_writable
    files = open_files('ftp:///ou*', host=host, port=port,
                       username=user, password=pw,
                       block_size=10000, cache_type=cache_type)
    assert len(files) == 1
    with files[0] as fo:
        assert fo.read(10) == b'hellohello'
        if isinstance(fo.cache, BytesCache):
            assert len(fo.cache.cache) == 10010
        assert fo.read(2) == b'he'
        assert fo.tell() == 12


def test_write_small(ftp_writable):
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host, port, user, pw)
    with fs.open('/out2', 'wb') as f:
        f.write(b'oi')
    assert fs.cat('/out2') == b'oi'


def test_with_url(ftp_writable):
    host, port, user, pw = ftp_writable
    fo = fsspec.open("ftp://{}:{}@{}:{}/out".format(user, pw, host, port), 'wb')
    with fo as f:
        f.write(b'hello')
    fo = fsspec.open("ftp://{}:{}@{}:{}/out".format(user, pw, host, port), 'rb')
    with fo as f:
        assert f.read() == b'hello'


@pytest.mark.parametrize('cache_type', ['bytes', 'mmap'])
def test_write_big(ftp_writable, cache_type):
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host, port, user, pw, block_size=1000,
                       cache_type=cache_type)
    fn = '/bigger'
    with fs.open(fn, 'wb') as f:
        f.write(b'o' * 500)
        assert not fs.exists(fn)
        f.write(b'o' * 1000)
        fs.invalidate_cache()
        assert fs.exists(fn)
        f.write(b'o' * 200)
        f.flush()
        assert f.buffer.tell() == 0

    assert fs.info(fn)['size'] == 1700
    assert fs.cat(fn) == b'o' * 1700


def test_transaction(ftp_writable):
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host, port, user, pw)
    fs.mkdir('/tmp')
    fn = '/tr'
    with fs.transaction:
        with fs.open(fn, 'wb') as f:
            f.write(b'not')
        assert not fs.exists(fn)
    assert fs.exists(fn)
    assert fs.cat(fn) == b'not'

    fs.rm(fn)
    assert not fs.exists(fn)




from __future__ import print_function, division, absolute_import

from io import BytesIO
from fsspec import AbstractFileSystem
import logging
logger = logging.Logger('fsspec.memoryfs')


class MemoryFileSystem(AbstractFileSystem):
    """A filesystem based on a dict of BytesIO objects"""
    store = {}  # global
    pseudo_dirs = []
    protocol = 'memory'
    root_marker = ''

    def ls(self, path, detail=False):
        if path in self.store:
            # there is a key with this exact name, but could also be directory
            out = [{'name': path,
                    'size': self.store[path].getbuffer().nbytes,
                    'type': 'file'}]
        else:
            out = []
        path = path.strip('/').lstrip('/')
        paths = set()
        for p2 in self.store:
            has_slash = "/" if p2.startswith('/') else ""
            p = p2.lstrip('/')
            if '/' in p:
                root = p.rsplit('/', 1)[0]
            else:
                root = ''
            if root == path:
                out.append({'name': has_slash + p,
                            'size': self.store[p2].getbuffer().nbytes,
                            'type': 'file'})
            elif path and all((a == b) for a, b
                              in zip(path.split('/'), p.strip('/').split('/'))):
                # implicit directory
                ppath = '/'.join(p.split('/')[:len(path.split('/')) + 1])
                if ppath not in paths:
                    out.append({'name': has_slash + ppath + '/',
                                'size': 0,
                                'type': 'directory'})
                    paths.add(ppath)
            elif all((a == b) for a, b
                     in zip(path.split('/'), [''] + p.strip('/').split('/'))):
                # root directory entry
                ppath = p.rstrip('/').split('/', 1)[0]
                if ppath not in paths:
                    out.append({'name': has_slash + ppath + '/',
                                'size': 0,
                                'type': 'directory'})
                    paths.add(ppath)
        for p2 in self.pseudo_dirs:
            if self._parent(p2).strip('/').rstrip('/') == path:
                out.append({'name': p2 + '/',
                            'size': 0,
                            'type': 'directory'})
        if detail:
            return out
        return sorted([f['name'] for f in out])

    def mkdir(self, path):
        path = path.rstrip('/')
        if path not in self.pseudo_dirs:
            self.pseudo_dirs.append(path)

    def rmdir(self, path):
        path = path.rstrip('/')
        if path in self.pseudo_dirs:
            if self.ls(path) == []:
                self.pseudo_dirs.remove(path)
            else:
                raise OSError("Directory %s not empty" % path)
        else:
            raise FileNotFoundError(path)

    def exists(self, path):
        return path in self.store

    def _open(self, path, mode='rb', **kwargs):
        """Make a file-like object

        Parameters
        ----------
        path: str
            identifier
        mode: str
            normally "rb", "wb" or "ab"
        """
        if mode in ['rb', 'ab', 'rb+']:
            if path in self.store:
                f = self.store[path]
                if mode == 'rb':
                    f.seek(0)
                else:
                    f.seek(0, 2)
                return f
            else:
                raise FileNotFoundError(path)
        if mode == 'wb':
            m = MemoryFile(self, path)
            if not self._intrans:
                m.commit()
            return m

    def copy(self, path1, path2, **kwargs):
        self.store[path2] = MemoryFile(self, path2,
                                       self.store[path1].getbuffer())

    def cat(self, path):
        return self.store[path].getvalue()

    def _rm(self, path):
        del self.store[path]

    def size(self, path):
        """Size in bytes of the file at path"""
        if path not in self.store:
            raise FileNotFoundError(path)
        return self.store[path].getbuffer().nbytes


class MemoryFile(BytesIO):
    """A BytesIO which can't close and works as a context manager

    Can initialise with data

    No need to provide fs, path if auto-committing (default)
    """

    def __init__(self, fs, path, data=None):
        self.fs = fs
        self.path = path
        if data:
            self.write(data)
            self.size = len(data)
            self.seek(0)

    def __enter__(self):
        return self

    def close(self):
        self.size = self.seek(0, 2)

    def discard(self):
        pass

    def commit(self):
        self.fs.store[self.path] = self



from __future__ import print_function, division, absolute_import

import re
import requests
from urllib.parse import urlparse
from fsspec import AbstractFileSystem
from fsspec.spec import AbstractBufferedFile
from fsspec.utils import tokenize, DEFAULT_BLOCK_SIZE

# https://stackoverflow.com/a/15926317/3821154
ex = re.compile(r"""<a\s+(?:[^>]*?\s+)?href=(["'])(.*?)\1""")
ex2 = re.compile(r"""(http[s]?://[-a-zA-Z0-9@:%_+.~#?&/=]+)""")


class HTTPFileSystem(AbstractFileSystem):
    """
    Simple File-System for fetching data via HTTP(S)

    ``ls()`` is implemented by loading the parent page and doing a regex
    match on the result. If simple_link=True, anything of the form
    "http(s)://server.com/stuff?thing=other"; otherwise only links within
    HTML href tags will be used.
    """
    sep = '/'

    def __init__(self, simple_links=True, block_size=None, same_scheme=True,
                 size_policy=None, **storage_options):
        """
        Parameters
        ----------
        block_size: int
            Blocks to read bytes; if 0, will default to raw requests file-like
            objects instead of HTTPFile instances
        simple_links: bool
            If True, will consider both HTML <a> tags and anything that looks
            like a URL; if False, will consider only the former.
        same_scheme: True
            When doing ls/glob, if this is True, only consider paths that have
            http/https matching the input URLs.
        size_policy: this argument is deprecated
        storage_options: key-value
            May be credentials, e.g., `{'auth': ('username', 'pword')}` or any
            other parameters passed on to requests
        """
        AbstractFileSystem.__init__(self)
        self.block_size = (block_size if block_size is not None
                           else DEFAULT_BLOCK_SIZE)
        self.simple_links = simple_links
        self.same_schema = same_scheme
        self.kwargs = storage_options
        self.session = requests.Session()

    @classmethod
    def _strip_protocol(cls, path):
        """ For HTTP, we always want to keep the full URL
        """
        return path

    def ls(self, url, detail=True):
        # ignoring URL-encoded arguments
        r = self.session.get(url, **self.kwargs)
        if self.simple_links:
            links = ex2.findall(r.text) + ex.findall(r.text)
        else:
            links = ex.findall(r.text)
        out = set()
        parts = urlparse(url)
        for l in links:
            if isinstance(l, tuple):
                l = l[1]
            if l.startswith('http'):
                if self.same_schema:
                    if l.split(':', 1)[0] == url.split(':', 1)[0]:
                        out.add(l)
                elif l.replace('https', 'http').startswith(
                        url.replace('https', 'http')):
                    # allowed to cross http <-> https
                    out.add(l)
            elif l.startswith('/') and len(l) > 1:
                out.add(parts.scheme + '://' + parts.netloc + l)
            else:
                if l not in ['..', '../']:
                    # Ignore FTP-like "parent"
                    out.add('/'.join([url.rstrip('/'), l.lstrip('/')]))
        if not out and url.endswith('/'):
            return self.ls(url.rstrip('/'), detail=True)
        if detail:
            return [{'name': u, 'size': None, 'type': 'directory'
                     if u.endswith('/') else 'file'} for u in out]
        else:
            return list(sorted(out))

    def cat(self, url):
        r = requests.get(url, **self.kwargs)
        r.raise_for_status()
        return r.content

    def mkdirs(self, url):
        """Make any intermediate directories to make path writable"""
        raise NotImplementedError

    def exists(self, path):
        kwargs = self.kwargs.copy()
        kwargs['stream'] = True
        try:
            r = self.session.get(path, **kwargs)
            r.close()
            return r.ok
        except requests.HTTPError:
            return False

    def _open(self, url, mode='rb', block_size=None, **kwargs):
        """Make a file-like object

        Parameters
        ----------
        url: str
            Full URL with protocol
        mode: string
            must be "rb"
        block_size: int or None
            Bytes to download in one request; use instance value if None. If
            zero, will return a streaming Requests file-like instance.
        kwargs: key-value
            Any other parameters, passed to requests calls
        """
        if mode != 'rb':
            raise NotImplementedError
        block_size = block_size if block_size is not None else self.block_size
        kw = self.kwargs.copy()
        kw.update(kwargs)
        kw.pop('autocommit', None)
        if block_size:
            return HTTPFile(self, url, self.session, block_size, **kw)
        else:
            kw['stream'] = True
            r = self.session.get(url, **kw)
            r.raise_for_status()
            r.raw.decode_content = True
            return r.raw

    def ukey(self, url):
        """Unique identifier; assume HTTP files are static, unchanging"""
        return tokenize(url, self.kwargs, self.protocol)

    def info(self, url, **kwargs):
        """Get info of URL

        Tries to access location via HEAD, and then GET methods, but does
        not fetch the data.

        It is possible that the server does not supply any size information, in
        which case size will be given as None (and certain operations on the
        corresponding file will not work).
        """
        size = False
        for policy in ['head', 'get']:
            try:
                size = file_size(url, self.session, policy, **self.kwargs)
                if size is not None:
                    break
            except Exception:
                pass
        else:
            # get failed, so conclude URL does not exist
            if size is False:
                raise FileNotFoundError(url)
        return {"name": url, "size": size, "type": "file"}


class HTTPFile(AbstractBufferedFile):
    """
    A file-like object pointing to a remove HTTP(S) resource

    Supports only reading, with read-ahead of a predermined block-size.

    In the case that the server does not supply the filesize, only reading of
    the complete file in one go is supported.

    Parameters
    ----------
    url: str
        Full URL of the remote resource, including the protocol
    session: requests.Session or None
        All calls will be made within this session, to avoid restarting
        connections where the server allows this
    block_size: int or None
        The amount of read-ahead to do, in bytes. Default is 5MB, or the value
        configured for the FileSystem creating this file
    size: None or int
        If given, this is the size of the file in bytes, and we don't attempt
        to call the server to find the value.
    kwargs: all other key-values are passed to requests calls.
    """

    def __init__(self, fs, url, session=None, block_size=None, mode='rb',
                 cache_type='bytes', size=None, **kwargs):
        if mode != 'rb':
            raise NotImplementedError('File mode not supported')
        self.url = url
        self.session = session if session is not None else requests.Session()
        if size is not None:
            self.details = {'name': url, 'size': size, 'type': 'file'}
        super().__init__(fs=fs, path=url, mode=mode, block_size=block_size,
                         cache_type=cache_type, **kwargs)
        self.cache.size = self.size or self.blocksize

    def read(self, length=-1):
        """Read bytes from file

        Parameters
        ----------
        length: int
            Read up to this many bytes. If negative, read all content to end of
            file. If the server has not supplied the filesize, attempting to
            read only part of the data will raise a ValueError.
        """
        if (
                (length < 0 and self.loc == 0) or  # explicit read all
                (length > (self.size or length)) or  # read more than there is
                (self.size and self.size < self.blocksize)  # all fits in one block anyway
        ):
            self._fetch_all()
        if self.size is None:
            if length < 0:
                self._fetch_all()
        else:
            length = min(self.size - self.loc, length)
        return super().read(length)

    def _fetch_all(self):
        """Read whole file in one shot, without caching

        This is only called when position is still at zero,
        and read() is called without a byte-count.
        """
        if not isinstance(self.cache, AllBytes):
            r = self.session.get(self.url, **self.kwargs)
            r.raise_for_status()
            out = r.content
            self.cache = AllBytes(out)
            self.size = len(out)

    def _fetch_range(self, start, end):
        """Download a block of data

        The expectation is that the server returns only the requested bytes,
        with HTTP code 206. If this is not the case, we first check the headers,
        and then stream the output - if the data size is bigger than we
        requested, an exception is raised.
        """
        kwargs = self.kwargs.copy()
        headers = kwargs.pop('headers', {})
        headers['Range'] = 'bytes=%i-%i' % (start, end - 1)
        r = self.session.get(self.url, headers=headers, stream=True, **kwargs)
        if r.status_code == 416:
            # range request outside file
            return b''
        r.raise_for_status()
        if r.status_code == 206:
            # partial content, as expected
            out = r.content
        elif 'Content-Length' in r.headers:
            cl = int(r.headers['Content-Length'])
            if cl <= end - start:
                # data size OK
                out = r.content
            else:
                raise ValueError('Got more bytes (%i) than requested (%i)' % (
                    cl, end - start))
        else:
            cl = 0
            out = []
            for chunk in r.iter_content(chunk_size=2 ** 20):
                # data size unknown, let's see if it goes too big
                if chunk:
                    out.append(chunk)
                    cl += len(chunk)
                    if cl > end - start:
                        raise ValueError(
                            'Got more bytes so far (>%i) than requested (%i)' % (
                                cl, end - start))
                else:
                    break
            out = b''.join(out)
        return out


def file_size(url, session=None, size_policy='head', **kwargs):
    """Call HEAD on the server to get file size

    Default operation is to explicitly allow redirects and use encoding
    'identity' (no compression) to get the true size of the target.
    """
    kwargs = kwargs.copy()
    ar = kwargs.pop('allow_redirects', True)
    head = kwargs.get('headers', {}).copy()
    head['Accept-Encoding'] = 'identity'
    session = session or requests.Session()
    if size_policy == 'head':
        r = session.head(url, allow_redirects=ar, **kwargs)
    elif size_policy == 'get':
        kwargs['stream'] = True
        r = session.get(url, allow_redirects=ar, **kwargs)
    else:
        raise TypeError('size_policy must be "head" or "get", got %s'
                        '' % size_policy)
    if 'Content-Length' in r.headers:
        return int(r.headers['Content-Length'])
    elif 'Content-Range' in r.headers:
        return int(r.headers['Content-Range'].split('/')[1])


class AllBytes(object):
    """Cache entire contents of a remote URL"""
    def __init__(self, data):
        self.data = data

    def _fetch(self, start, end):
        return self.data[start:end]



from __future__ import print_function
import os
import stat
from errno import ENOENT, EIO
from fuse import Operations, FuseOSError
import threading
import time
from fuse import FUSE


class FUSEr(Operations):

    def __init__(self, fs, path):
        self.fs = fs
        self.cache = {}
        self.root = path.rstrip('/') + '/'
        self.counter = 0
    
    def getattr(self, path, fh=None):
        path = ''.join([self.root, path.lstrip('/')]).rstrip('/')
        try:
            info = self.fs.info(path)
        except FileNotFoundError:
            raise FuseOSError(ENOENT)
        data = {'st_uid': 1000, 'st_gid': 1000}
        perm = 0o777

        if info['type'] != 'file':
            data['st_mode'] = (stat.S_IFDIR | perm)
            data['st_size'] = 0
            data['st_blksize'] = 0
        else:
            data['st_mode'] = (stat.S_IFREG | perm)
            data['st_size'] = info['size']
            data['st_blksize'] = 5 * 2**20
            data['st_nlink'] = 1
        data['st_atime'] = time.time()
        data['st_ctime'] = time.time()
        data['st_mtime'] = time.time()
        return data

    def readdir(self, path, fh):
        path = ''.join([self.root, path.lstrip('/')])
        files = self.fs.ls(path, False)
        files = [os.path.basename(f.rstrip('/')) for f in files]
        return ['.', '..'] + files
    
    def mkdir(self, path, mode):
        path = ''.join([self.root, path.lstrip('/')])
        self.fs.mkdir(path)
        return 0
    
    def rmdir(self, path):
        path = ''.join([self.root, path.lstrip('/')])
        self.fs.rmdir(path)
        return 0
    
    def read(self, path, size, offset, fh):
        f = self.cache[fh]
        f.seek(offset)
        out = f.read(size)
        return out
    
    def write(self, path, data, offset, fh):
        f = self.cache[fh]
        f.write(data)
        return len(data)
    
    def create(self, path, flags, fi=None):
        fn = ''.join([self.root, path.lstrip('/')])
        f = self.fs.open(fn, 'wb')
        self.cache[self.counter] = f
        self.counter += 1
        return self.counter - 1
    
    def open(self, path, flags):
        fn = ''.join([self.root, path.lstrip('/')])
        if flags % 2 == 0:
            # read
            mode = 'rb'
        else:
            # write/create
            mode = 'wb'
        self.cache[self.counter] = self.fs.open(fn, mode)
        self.counter += 1
        return self.counter - 1

    def truncate(self, path, length, fh=None):
        fn = ''.join([self.root, path.lstrip('/')])
        if length != 0:
            raise NotImplementedError
        # maybe should be no-op since open with write sets size to zero anyway
        self.fs.touch(fn)
    
    def unlink(self, path):
        fn = ''.join([self.root, path.lstrip('/')])
        try:
            self.fs.rm(fn, False)
        except (IOError, FileNotFoundError):
            raise FuseOSError(EIO)
    
    def release(self, path, fh):
        try:
            if fh in self.cache:
                f = self.cache[fh]
                f.close()
                self.cache.pop(fh)
        except Exception as e:
            print(e)
        return 0
    
    def chmod(self, path, mode):
        raise NotImplementedError


def run(fs, path, mount_point, foreground=True, threads=False):
    """ Mount stuff in a local directory

    This uses fusepy to make it appear as if a given path on an fsspec
    instance is in fact resident within the local file-system.

    This requires that fusepy by installed, and that FUSE be available on
    the system (typically requiring a package to be installed with
    apt, yum, brew, etc.).

    Parameters
    ----------
    fs: file-system instance
        From one of the compatible implementations
    path: str
        Location on that file-system to regard as the root directory to
        mount. Note that you typically should include the terminating "/"
        character.
    mount_point: str
        An empty directory on the local file-system where the contents of
        the remote path will appear
    foreground: bool
        Whether or not calling this function will block. Operation will
        typically be more stable if True.
    threads: bool
        Whether or not to create threads when responding to file operations
        within the mounter directory. Operation will typically be more
        stable if False.

    """
    func = lambda: FUSE(FUSEr(fs, path),
                        mount_point, nothreads=not threads, foreground=True)
    if foreground is False:
        th = threading.Thread(target=func)
        th.daemon = True
        th.start()
        return th
    else:  # pragma: no cover
        try:
            func()
        except KeyboardInterrupt:
            pass



"""Tests the spec, using memoryfs"""

import os
import pickle
from fsspec.implementations.memory import MemoryFileSystem, MemoryFile


def test_idempotent():
    MemoryFileSystem.clear_instance_cache()
    fs = MemoryFileSystem()
    fs2 = MemoryFileSystem()
    assert fs is fs2
    assert MemoryFileSystem.current() is fs2
    fs2 = MemoryFileSystem(do_cache=False)
    assert fs is not fs2

    assert hash(fs) == hash(fs2)
    assert fs == fs2

    MemoryFileSystem.clear_instance_cache()
    assert not MemoryFileSystem._cache

    fs2 = MemoryFileSystem().current()
    assert fs == fs2


def test_pickle():
    fs = MemoryFileSystem()
    fs2 = pickle.loads(pickle.dumps(fs))
    assert fs == fs2


def test_class_methods():
    assert MemoryFileSystem._strip_protocol('memory:stuff') == "stuff"
    assert MemoryFileSystem._strip_protocol('memory://stuff') == "stuff"
    assert MemoryFileSystem._strip_protocol('stuff') == "stuff"
    assert MemoryFileSystem._strip_protocol('other://stuff') == "other://stuff"

    assert MemoryFileSystem._get_kwargs_from_urls("memory://user@thing") == {}


def test_get_put(tmpdir):
    tmpdir = str(tmpdir)
    fn = os.path.join(tmpdir, 'one')
    open(fn, 'wb').write(b'one')
    os.mkdir(os.path.join(tmpdir, 'dir'))
    fn2 = os.path.join(tmpdir, 'dir', 'two')
    open(fn2, 'wb').write(b'two')

    fs = MemoryFileSystem()
    fs.put(fn, '/afile')
    assert fs.cat('/afile') == b'one'

    fs.store['/bfile'] = MemoryFile(fs, '/bfile', b'data')
    fn3 = os.path.join(tmpdir, 'three')
    fs.get('/bfile', fn3)
    assert open(fn3, 'rb').read() == b'data'

    fs.put(tmpdir, '/more', recursive=True)
    assert fs.find('/more') == ['/more/dir/two', '/more/one', '/more/three']

    for f in [fn, fn2, fn3]:
        os.remove(f)
    os.rmdir(os.path.join(tmpdir, 'dir'))

    fs.get('/more/', tmpdir + '/', recursive=True)
    assert open(fn3, 'rb').read() == b'data'
    assert open(fn, 'rb').read() == b'one'


def test_du():
    fs = MemoryFileSystem()
    fs.store = {
        '/dir/afile': MemoryFile(fs, '/afile', b'a'),
        '/dir/dirb/afile': MemoryFile(fs, '/afile', b'bb'),
        '/dir/dirb/bfile': MemoryFile(fs, '/afile', b'ccc'),
    }
    assert fs.du('/dir') == 6
    assert fs.du('/dir', total=False)['/dir/dirb/afile'] == 2
    assert fs.du('/dir', maxdepth=0) == 1


def test_head_tail():
    fs = MemoryFileSystem()
    with fs.open('/myfile', 'wb') as f:
        f.write(b'I had a nice big cabbage')
    assert fs.head('/myfile', 5) == b'I had'
    assert fs.tail('/myfile', 7) == b'cabbage'


def test_move():
    fs = MemoryFileSystem()
    with fs.open('/myfile', 'wb') as f:
        f.write(b'I had a nice big cabbage')
    fs.move('/myfile', '/otherfile')
    assert not fs.exists('/myfile')
    assert fs.info('/otherfile')
    assert isinstance(fs.ukey('/otherfile'), str)


def test_read_block_delimiter():
    fs = MemoryFileSystem()
    with fs.open('/myfile', 'wb') as f:
        f.write(b'some\n'
                b'lines\n'
                b'of\n'
                b'text')
    assert fs.read_block('/myfile', 0, 2, b'\n') == b'some\n'
    assert fs.read_block('/myfile', 2, 6, b'\n') == b'lines\n'
    assert fs.read_block('/myfile', 6, 2, b'\n') == b''
    assert fs.read_block('/myfile', 2, 9, b'\n') == b'lines\nof\n'
    assert fs.read_block('/myfile', 12, 6, b'\n') == b'text'
    assert fs.read_block('/myfile', 0, None) == fs.cat('/myfile')



from ftplib import FTP, Error
import uuid
from ..spec import AbstractBufferedFile, AbstractFileSystem
from ..utils import infer_storage_options


class FTPFileSystem(AbstractFileSystem):
    """A filesystem over classic """
    root_marker = '/'

    def __init__(self, host, port=21, username=None, password=None,
                 acct=None, block_size=None, tempdir='/tmp', **kwargs):
        """
        You can use _get_kwargs_from_urls to get some kwargs from
        a reasonable FTP url.

        Authentication will be anonymous if username/password are not
        given.

        Parameters
        ----------
        host: str
            The remote server name/ip to connect to
        port: int
            Port to connect with
        username: str or None
            If authenticating, the user's identifier
        password: str of None
            User's password on the server, if using
        acct: str or None
            Some servers also need an "account" string for auth
        block_size: int or None
            If given, the read-ahead or write buffer size.
        tempdir: str
            Directory on remote to put temporary files when in a transaction
        """
        super(FTPFileSystem, self).__init__(**kwargs)
        self.host = host
        self.port = port
        self.tempdir = tempdir
        self.cred = username, password, acct
        if block_size is not None:
            self.blocksize = block_size
        self._connect()

    def _connect(self):
        self.ftp = FTP()
        self.ftp.connect(self.host, self.port)
        self.ftp.login(*self.cred)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('ftp')
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._connect()

    @classmethod
    def _strip_protocol(cls, path):
        return '/' + infer_storage_options(path)['path'].lstrip('/').rstrip('/')

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        out = infer_storage_options(urlpath)
        out.pop('path', None)
        out.pop('protocol', None)
        return out

    def invalidate_cache(self, path=None):
        if path is not None:
            self.dircache.pop(path, None)
        else:
            self.dircache.clear()

    def ls(self, path, detail=True):
        path = self._strip_protocol(path)
        out = []
        if path not in self.dircache:
            try:
                out = [(fn, details) for (fn, details) in self.ftp.mlsd(path)
                       if fn not in ['.', '..']
                       and details['type'] not in ['pdir', 'cdir']]
                for fn, details in out:
                    if path == '/':
                        path = ''  # just for forming the names, below
                    details['name'] = '/'.join([path, fn.lstrip('/')])
                    if details['type'] == 'file':
                        details['size'] = int(details['size'])
                    else:
                        details['size'] = 0
                self.dircache[path] = out
            except Error:
                try:
                    info = self.info(path)
                    if info['type'] == 'file':
                        out = [(path, info)]
                except (Error, IndexError):
                    raise FileNotFoundError
        files = self.dircache.get(path, out)
        if not detail:
            return sorted([fn for fn, details in files])
        return [details for fn, details in files]

    def info(self, path, **kwargs):
        # implement with direct method
        path = self._strip_protocol(path)
        files = self.ls(self._parent(path), True)
        return [f for f in files if f['name'] == path][0]

    def _open(self, path, mode='rb', block_size=None, autocommit=True,
              **kwargs):
        path = self._strip_protocol(path)
        block_size = block_size or self.blocksize
        return FTPFile(self, path, mode=mode, block_size=block_size,
                       tempdir=self.tempdir, autocommit=autocommit)

    def _rm(self, path):
        path = self._strip_protocol(path)
        self.ftp.delete(path)
        self.invalidate_cache(path.rsplit('/', 1)[0])

    def mkdir(self, path, **kwargs):
        path = self._strip_protocol(path)
        self.ftp.mkd(path)

    def rmdir(self, path):
        path = self._strip_protocol(path)
        self.ftp.rmd(path)

    def mv(self, path1, path2, **kwargs):
        path1 = self._strip_protocol(path1)
        path2 = self._strip_protocol(path2)
        self.ftp.rename(path1, path2)
        self.invalidate_cache(self._parent(path1))
        self.invalidate_cache(self._parent(path2))

    def __del__(self):
        self.ftp.close()


class TransferDone(Exception):
    """Internal exception to break out of transfer"""
    pass


class FTPFile(AbstractBufferedFile):
    """Interact with a remote FTP file with read/write buffering"""

    def __init__(self, fs, path, **kwargs):
        super().__init__(fs, path, **kwargs)
        if kwargs.get('autocommit', False) is False:
            self.target = self.path
            self.path = '/'.join([kwargs['tempdir'], str(uuid.uuid4())])

    def commit(self):
        self.fs.mv(self.path, self.target)

    def discard(self):
        self.fs.rm(self.path)

    def _fetch_range(self, start, end):
        """Get bytes between given byte limits

        Implemented by raising an exception in the fetch callback when the
        number of bytes received reaches the requested amount.

        With fail if the server does not respect the REST command on
        retrieve requests.
        """
        out = []
        total = [0]

        def callback(x):
            total[0] += len(x)
            if total[0] > end - start:
                out.append(x[:(end - start) - total[0]])
                self.fs.ftp.abort()
                raise TransferDone
            else:
                out.append(x)

            if total[0] == end - start:
                raise TransferDone

        try:
            self.fs.ftp.retrbinary('RETR %s' % self.path, blocksize=2**16,
                                   rest=start, callback=callback)
        except TransferDone:
            self.fs.ftp.abort()
            self.fs.ftp.voidresp()
        return b''.join(out)

    def _upload_chunk(self, final=False):
        self.buffer.seek(0)
        self.fs.ftp.storbinary("STOR " + self.path, self.buffer,
                               blocksize=2**16, rest=self.offset)
        return True



import pytest
import fsspec


@pytest.fixture()
def m():
    m = fsspec.filesystem('memory')
    m.store.clear()
    try:
        yield m
    finally:
        m.store.clear()


def test_1(m):
    m.touch('/somefile')  # NB: is found with or without initial /
    m.touch('afiles/and/anothers')
    assert m.find('') == ['afiles/and/anothers', 'somefile']
    assert list(m.get_mapper('')) == ['afiles/and/anothers', 'somefile']


def test_ls(m):
    m.touch('/dir/afile')
    m.touch('/dir/dir1/bfile')
    m.touch('/dir/dir1/cfile')

    assert m.ls('/', False) == ['/dir/']
    assert m.ls('/dir', False) == ['/dir/afile', '/dir/dir1/']
    assert m.ls('/dir', True)[0]['type'] == 'file'
    assert m.ls('/dir', True)[1]['type'] == 'directory'

    assert len(m.ls('/dir/dir1')) == 2



import zipfile
from contextlib import contextmanager
import os
import pickle
import pytest
import sys
import tempfile
import fsspec


@contextmanager
def tempzip(data={}):
    f = tempfile.mkstemp(suffix='zip')[1]
    with zipfile.ZipFile(f, mode='w') as z:
        for k, v in data.items():
            z.writestr(k, v)
    try:
        yield f
    finally:
        try:
            os.remove(f)
        except (IOError, OSError):
            pass


data = {'a': b'',
        'b': b'hello',
        'deeply/nested/path': b"stuff"}


def test_empty():
    with tempzip() as z:
        fs = fsspec.get_filesystem_class('zip')(fo=z)
        assert fs.find('') == []


@pytest.mark.skipif(sys.version_info.minor < 6, reason="zip-info odd on py35")
def test_mapping():
    with tempzip(data) as z:
        fs = fsspec.get_filesystem_class('zip')(fo=z)
        m = fs.get_mapper('')
        assert list(m) == ['a', 'b', 'deeply/nested/path']
        assert m['b'] == data['b']


def test_pickle():
    with tempzip(data) as z:
        fs = fsspec.get_filesystem_class('zip')(fo=z)
        fs2 = pickle.loads(pickle.dumps(fs))
        assert fs2.cat('b') == b'hello'



from __future__ import print_function, division, absolute_import

import zipfile
from fsspec import AbstractFileSystem, open_files
from fsspec.utils import tokenize, DEFAULT_BLOCK_SIZE


class ZipFileSystem(AbstractFileSystem):
    """Read contents of ZIP archive as a file-system

    Keeps file object open while instance lives.

    This class is pickleable, but not necessarily thread-safe
    """
    root_marker = ""

    def __init__(self, fo='', mode='r', **storage_options):
        """
        Parameters
        ----------
        fo: str or file-like
            Contains ZIP, and must exist. If a str, will fetch file using
            `open_files()`, which must return one file exactly.
        mode: str
            Currently, only 'r' accepted
        storage_options: key-value
            May be credentials, e.g., `{'auth': ('username', 'pword')}` or any
            other parameters for requests
        """
        AbstractFileSystem.__init__(self)
        if mode != 'r':
            raise ValueError("Only read from zip files accepted")
        self.in_fo = fo
        if isinstance(fo, str):
            files = open_files(fo)
            if len(files) != 1:
                raise ValueError('Path "{}" did not resolve to exactly'
                                 'one file: "{}"'.format(fo, files))
            fo = files[0]
        self.fo = fo.__enter__()  # the whole instance is a context
        self.zip = zipfile.ZipFile(self.fo)
        self.block_size = storage_options.pop('block_size', DEFAULT_BLOCK_SIZE)
        self.kwargs = storage_options
        self.dir_cache = None

    @classmethod
    def _strip_protocol(cls, path):
        # zip file paths are always relative to the archive root
        return super()._strip_protocol(path).lstrip('/')

    def _get_dirs(self):
        if self.dir_cache is None:
            files = self.zip.infolist()
            self.dir_cache = {}
            for z in files:
                f = {s: getattr(z, s) for s in zipfile.ZipInfo.__slots__}
                f.update({'name': z.filename, 'size': z.file_size,
                          'type': ('directory' if z.is_dir() else 'file')})
                self.dir_cache[f['name']] = f

    def ls(self, path, detail=False):
        self._get_dirs()
        paths = {}
        for p, f in self.dir_cache.items():
            p = p.rstrip('/')
            if '/' in p:
                root = p.rsplit('/', 1)[0]
            else:
                root = ''
            if root == path.rstrip('/'):
                paths[p] = f
            elif path and all(
                    (a == b) for a, b
                    in zip(path.split('/'), p.strip('/').split('/'))):
                # implicit directory
                ppath = '/'.join(p.split('/')[:len(path.split('/')) + 1])
                if ppath not in paths:
                    out = {'name': ppath + '/', 'size': 0, 'type': 'directory'}
                    paths[ppath] = out

            elif all((a == b) for a, b
                     in zip(path.split('/'), [''] + p.strip('/').split('/'))):
                # root directory entry
                ppath = p.rstrip('/').split('/', 1)[0]
                if ppath not in paths:
                    out = {'name': ppath + '/', 'size': 0, 'type': 'directory'}
                    paths[ppath] = out
        out = list(paths.values())
        if detail:
            return out
        else:
            return list(sorted(f['name'] for f in out))

    def __reduce_ex__(self, protocol):
        return ZipFileSystem, (self.in_fo, )

    def cat(self, path):
        return self.zip.read(path)

    def _open(self, path, mode='rb', **kwargs):
        if mode != 'rb':
            raise NotImplementedError
        info = self.info(path)
        out = self.zip.open(path, 'r')
        out.size = info['size']
        out.name = info['name']
        return out

    def ukey(self, path):
        return tokenize(path, self.kwargs, self.protocol)



import importlib
from distutils.version import LooseVersion

__all__ = ['registry', 'get_filesystem_class', 'default']

# mapping protocol: implementation class object
registry = {}
default = 'file'

# protocols mapped to the class which implements them. This dict can
# be dynamically updated.
known_implementations = {
    'file': {'class': 'fsspec.implementations.local.LocalFileSystem'},
    'memory': {'class': 'fsspec.implementations.memory.MemoryFileSystem'},
    'http': {'class': 'fsspec.implementations.http.HTTPFileSystem',
             'err': 'HTTPFileSystem requires "requests" to be installed'},
    'https': {'class': 'fsspec.implementations.http.HTTPFileSystem',
              'err': 'HTTPFileSystem requires "requests" to be installed'},
    'zip': {'class': 'fsspec.implementations.zip.ZipFileSystem'},
    'gcs': {'class': 'gcsfs.GCSFileSystem',
            'err': 'Please install gcsfs to access Google Storage'},
    'gs': {'class': 'gcsfs.GCSFileSystem',
           'err': 'Please install gcsfs to access Google Storage'},
    'sftp': {'class': 'fsspec.implementations.sftp.SFTPFileSystem',
             'err': 'SFTPFileSystem requires "paramiko" to be installed'},
    'ssh': {'class': 'fsspec.implementations.sftp.SFTPFileSystem',
            'err': 'SFTPFileSystem requires "paramiko" to be installed'},
    'ftp': {'class': 'fsspec.implementations.ftp.FTPFileSystem'},
    'hdfs': {'class': 'fsspec.implementations.hdfs.PyArrowHDFS',
             'err': 'pyarrow and local java libraries required for HDFS'},
    'webhdfs': {'class': 'fsspec.implementations.webhdfs.WebHDFS',
                'err': 'webHDFS access requires "requests" to be installed'},
    's3': {'class': 's3fs.S3FileSystem',
           'err': 'Install s3fs to access S3'},
    'cached': {'class': 'fsspec.implementations.cached.CachingFileSystem'},
    'blockcache': {'class': 'fsspec.implementations.cached.CachingFileSystem'},
    'filecache': {
        'class': 'fsspec.implementations.cached.WholeFileCacheFileSystem'},
    'dask': {'class': 'fsspec.implementations.dask.DaskWorkerFileSystem',
             'err': 'Install dask distributed to access worker file system'}
}

minversions = {
    's3fs': LooseVersion("0.3.0"),
    "gcsfs": LooseVersion("0.3.0"),
}


def get_filesystem_class(protocol):
    """Fetch named protocol implementation from the registry

    The dict ``known_implementations`` maps protocol names to the locations
    of classes implementing the corresponding file-system. When used for the
    first time, appropriate imports will happen and the class will be placed in
    the registry. All subsequent calls will fetch directly from the registry.

    Some protocol implementations require additional dependencies, and so the
    import may fail. In this case, the string in the "err" field of the
    ``known_implementations`` will be given as the error message.
    """
    if protocol is None:
        protocol = default

    if protocol not in registry:
        if protocol not in known_implementations:
            raise ValueError("Protocol not known: %s" % protocol)
        bit = known_implementations[protocol]
        mod, name = bit['class'].rsplit('.', 1)
        minversion = minversions.get(mod, None)
        err = None
        try:
            mod = importlib.import_module(mod)
        except ImportError:
            err = ImportError(bit['err'])

        except Exception as e:
            err = e
        if err is not None:
            raise RuntimeError(str(err))

        if minversion:
            version = getattr(mod, "__version__", None)
            if version and LooseVersion(version) < minversion:
                raise RuntimeError("'{}={}' is installed, but version '{}' or "
                                   "higher is required".format(mod.__name__,
                                                               version,
                                                               minversion))
        registry[protocol] = getattr(mod, name)
    cls = registry[protocol]
    if getattr(cls, 'protocol', None) in ('abstract', None):
        cls.protocol = protocol

    return cls


def filesystem(protocol, **storage_options):
    """Instantiate filesystems for given protocol and arguments

    ``storage_options`` are specific to the protocol being chosen, and are
    passed directly to the class.
    """
    cls = get_filesystem_class(protocol)
    return cls(**storage_options)



import pytest
import fsspec

pytest.importorskip("distributed")


@pytest.fixture()
def cli(tmpdir):
    import dask.distributed
    client = dask.distributed.Client(n_workers=1)

    def setup():
        m = fsspec.filesystem("memory")
        with m.open('afile', 'wb') as f:
            f.write(b'data')

    client.run(setup)
    try:
        yield client
    finally:
        client.close()


def test_basic(cli):

    fs = fsspec.filesystem('dask', remote_protocol='memory')
    assert fs.ls('') == ['afile']
    assert fs.cat('afile') == b'data'



import pathlib

import pytest

import fsspec.core
from fsspec.compression import compr, register_compression
from fsspec.utils import compressions, infer_compression


def test_infer_custom_compression():
    """Inferred compression gets values from fsspec.compression.compr."""
    assert infer_compression("fn.zip") == "zip"
    assert infer_compression("fn.gz") == "gzip"
    assert infer_compression("fn.unknown") is None
    assert infer_compression("fn.test_custom") is None
    assert infer_compression("fn.tst") is None

    register_compression("test_custom", lambda f, **kwargs: f, "tst")

    try:
        assert infer_compression("fn.zip") == "zip"
        assert infer_compression("fn.gz") == "gzip"
        assert infer_compression("fn.unknown") is None
        assert infer_compression("fn.test_custom") is None
        assert infer_compression("fn.tst") == "test_custom"

        # Duplicate registration in name or extension raises a value error.
        with pytest.raises(ValueError):
            register_compression("test_custom", lambda f, **kwargs: f, "tst")

        with pytest.raises(ValueError):
            register_compression("test_conflicting", lambda f, **kwargs: f, "tst")
        assert "test_conflicting" not in compr

        # ...but can be forced.
        register_compression(
            "test_conflicting", lambda f, **kwargs: f, "tst", force=True
        )
        assert infer_compression("fn.zip") == "zip"
        assert infer_compression("fn.gz") == "gzip"
        assert infer_compression("fn.unknown") is None
        assert infer_compression("fn.test_custom") is None
        assert infer_compression("fn.tst") == "test_conflicting"

    finally:
        del compr["test_custom"]
        del compr["test_conflicting"]
        del compressions["tst"]


def test_lzma_compression_name():
    pytest.importorskip("lzma")
    assert infer_compression("fn.xz") == "xz"


def test_lz4_compression(tmp_path):
    # type: (pathlib.Path) -> None
    """Infer lz4 compression for .lz4 files if lz4 is available."""

    lz4 = pytest.importorskip("lz4")

    tmp_path.mkdir(exist_ok=True)

    tdat = "foobar" * 100

    with fsspec.core.open(
        str(tmp_path / "out.lz4"), mode="wt", compression="infer"
    ) as outfile:
        outfile.write(tdat)

    compressed = (tmp_path / "out.lz4").open("rb").read()
    assert lz4.frame.decompress(compressed).decode() == tdat

    with fsspec.core.open(
        str(tmp_path / "out.lz4"), mode="rt", compression="infer"
    ) as infile:
        assert infile.read() == tdat

    with fsspec.core.open(
        str(tmp_path / "out.lz4"), mode="rt", compression="lz4"
    ) as infile:
        assert infile.read() == tdat


def test_zstd_compression(tmp_path):
    # type: (pathlib.Path) -> None
    """Infer zstd compression for .zst files if zstandard is available."""

    zstd = pytest.importorskip("zstandard")

    tmp_path.mkdir(exist_ok=True)

    tdat = "foobar" * 100

    with fsspec.core.open(
        str(tmp_path / "out.zst"), mode="wt", compression="infer"
    ) as outfile:
        outfile.write(tdat)

    compressed = (tmp_path / "out.zst").open("rb").read()
    assert zstd.ZstdDecompressor().decompress(compressed, len(tdat)).decode() == tdat

    with fsspec.core.open(
        str(tmp_path / "out.zst"), mode="rt", compression="infer"
    ) as infile:
        assert infile.read() == tdat

    with fsspec.core.open(
        str(tmp_path / "out.zst"), mode="rt", compression="zstd"
    ) as infile:
        assert infile.read() == tdat


def test_snappy_compression(tmp_path):
    """No registered compression for snappy, but can be specified."""
    # type: (pathlib.Path) -> None

    snappy = pytest.importorskip("snappy")

    tmp_path.mkdir(exist_ok=True)

    tdat = "foobar" * 100

    # Snappy isn't inferred.
    with fsspec.core.open(
        str(tmp_path / "out.snappy"), mode="wt", compression="infer"
    ) as outfile:
        outfile.write(tdat)
    assert (tmp_path / "out.snappy").open("rb").read().decode() == tdat

    # but can be specified.
    with fsspec.core.open(
        str(tmp_path / "out.snappy"), mode="wt", compression="snappy"
    ) as outfile:
        outfile.write(tdat)

    compressed = (tmp_path / "out.snappy").open("rb").read()
    assert snappy.StreamDecompressor().decompress(compressed).decode() == tdat

    with fsspec.core.open(
        str(tmp_path / "out.snappy"), mode="rb", compression="infer"
    ) as infile:
        assert infile.read() == compressed

    with fsspec.core.open(
        str(tmp_path / "out.snappy"), mode="rt", compression="snappy"
    ) as infile:
        assert infile.read() == tdat



import os
import pytest
import pickle
from fsspec.core import _expand_paths, OpenFile, caches, get_compression
from fsspec.implementations.tests.test_memory import m


@pytest.mark.parametrize('path, name_function, num, out', [
    [['apath'], None, 1, ['apath']],
    ['apath.*.csv', None, 1, ['apath.0.csv']],
    ['apath.*.csv', None, 2, ['apath.0.csv', 'apath.1.csv']],
    ['a*', lambda x: 'abc'[x], 2, ['aa', 'ab']]
])
def test_expand_paths(path, name_function, num, out):
    assert _expand_paths(path, name_function, num) == out


def test_expand_error():
    with pytest.raises(ValueError):
        _expand_paths("*.*", None, 1)


def test_openfile_api(m):
    m.open('somepath', 'wb').write(b'data')
    of = OpenFile(m, 'somepath')
    assert str(of) == "<OpenFile 'somepath'>"
    f = of.open()
    assert f.read() == b'data'
    f.close()
    with OpenFile(m, 'somepath', mode='rt') as f:
        f.read() == 'data'


# For test_cache_pickleable(). Functions are only picklable if they are defined
# at the top-level of a module
def _fetcher(start, end):
    return b'0' * (end - start)


@pytest.mark.parametrize('Cache_imp', caches.values())
def test_cache_empty_file(Cache_imp):
    blocksize = 5
    size = 0
    cache = Cache_imp(blocksize, _fetcher, size)
    assert cache._fetch(0, 0) == b''


@pytest.mark.parametrize('Cache_imp', caches.values())
def test_cache_pickleable(Cache_imp):
    blocksize = 5
    size = 100
    cache = Cache_imp(blocksize, _fetcher, size)
    cache._fetch(0, 5)  # fill in cache
    unpickled = pickle.loads(pickle.dumps(cache))
    assert isinstance(unpickled, Cache_imp)
    assert unpickled.blocksize == blocksize
    assert unpickled.size == size
    assert unpickled._fetch(0, 10) == b'0' * 10


def test_xz_lzma_compressions():
    pytest.importorskip("lzma")
    # Ensure that both 'xz' and 'lzma' compression names can be parsed
    assert get_compression('some_file.xz', 'infer') == 'xz'
    assert get_compression('some_file.xz', 'xz') == 'xz'
    assert get_compression('some_file.xz', 'lzma') == 'lzma'



from __future__ import print_function, division, absolute_import

import io
import os
import logging
from .compression import compr
from .utils import (infer_compression, build_name_function,
                    update_storage_options, stringify_path)
from .registry import get_filesystem_class
logger = logging.getLogger('fsspec')


class OpenFile(object):
    """
    File-like object to be used in a context

    Can layer (buffered) text-mode and compression over any file-system, which
    are typically binary-only.

    These instances are safe to serialize, as the low-level file object
    is not created until invoked using `with`.

    Parameters
    ----------
    fs: FileSystem
        The file system to use for opening the file. Should match the interface
        of ``dask.bytes.local.LocalFileSystem``.
    path: str
        Location to open
    mode: str like 'rb', optional
        Mode of the opened file
    compression: str or None, optional
        Compression to apply
    encoding: str or None, optional
        The encoding to use if opened in text mode.
    errors: str or None, optional
        How to handle encoding errors if opened in text mode.
    newline: None or str
        Passed to TextIOWrapper in text mode, how to handle line endings.
    """
    def __init__(self, fs, path, mode='rb', compression=None, encoding=None,
                 errors=None, newline=None):
        self.fs = fs
        self.path = path
        self.mode = mode
        self.compression = get_compression(path, compression)
        self.encoding = encoding
        self.errors = errors
        self.newline = newline
        self.fobjects = []

    def __reduce__(self):
        return (OpenFile, (self.fs, self.path, self.mode, self.compression,
                           self.encoding, self.errors))

    def __repr__(self):
        return "<OpenFile '{}'>".format(self.path)

    def __fspath__(self):
        return self.path

    def __enter__(self):
        mode = self.mode.replace('t', '').replace('b', '') + 'b'

        f = self.fs.open(self.path, mode=mode)

        self.fobjects = [f]

        if self.compression is not None:
            compress = compr[self.compression]
            f = compress(f, mode=mode[0])
            self.fobjects.append(f)

        if 'b' not in self.mode:
            # assume, for example, that 'r' is equivalent to 'rt' as in builtin
            f = io.TextIOWrapper(f, encoding=self.encoding,
                                 errors=self.errors, newline=self.newline)
            self.fobjects.append(f)

        return self.fobjects[-1]

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        self.close()

    def open(self):
        """Materialise this as a real open file without context

        The file should be explicitly closed to avoid enclosed open file
        instances persisting
        """
        return self.__enter__()

    def close(self):
        """Close all encapsulated file objects"""
        for f in reversed(self.fobjects):
            if 'r' not in self.mode and not f.closed:
                f.flush()
            f.close()
        self.fobjects = []


def open_files(urlpath, mode='rb', compression=None, encoding='utf8',
               errors=None, name_function=None, num=1, protocol=None,
               newline=None, **kwargs):
    """ Given a path or paths, return a list of ``OpenFile`` objects.

    For writing, a str path must contain the "*" character, which will be filled
    in by increasing numbers, e.g., "part*" ->  "part1", "part2" if num=2.

    For either reading or writing, can instead provide explicit list of paths.

    Parameters
    ----------
    urlpath: string or list
        Absolute or relative filepath(s). Prefix with a protocol like ``s3://``
        to read from alternative filesystems. To read from multiple files you
        can pass a globstring or a list of paths, with the caveat that they
        must all have the same protocol.
    mode: 'rb', 'wt', etc.
    compression: string
        Compression to use.  See ``dask.bytes.compression.files`` for options.
    encoding: str
        For text mode only
    errors: None or str
        Passed to TextIOWrapper in text mode
    name_function: function or None
        if opening a set of files for writing, those files do not yet exist,
        so we need to generate their names by formatting the urlpath for
        each sequence number
    num: int [1]
        if writing mode, number of files we expect to create (passed to
        name+function)
    protocol: str or None
        If given, overrides the protocol found in the URL.
    newline: bytes or None
        Used for line terminator in text mode. If None, uses system default;
        if blank, uses no translation.
    **kwargs: dict
        Extra options that make sense to a particular storage connection, e.g.
        host, port, username, password, etc.

    Examples
    --------
    >>> files = open_files('2015-*-*.csv')  # doctest: +SKIP
    >>> files = open_files('s3://bucket/2015-*-*.csv.gz', compression='gzip')  # doctest: +SKIP

    Returns
    -------
    List of ``OpenFile`` objects.
    """
    fs, fs_token, paths = get_fs_token_paths(urlpath, mode, num=num,
                                             name_function=name_function,
                                             storage_options=kwargs,
                                             protocol=protocol)
    return [OpenFile(fs, path, mode=mode, compression=compression,
                     encoding=encoding, errors=errors, newline=newline)
            for path in paths]


def open(urlpath, mode='rb', compression=None, encoding='utf8',
         errors=None, protocol=None, newline=None, **kwargs):
    """ Given a path or paths, return one ``OpenFile`` object.

    Parameters
    ----------
    urlpath: string or list
        Absolute or relative filepath. Prefix with a protocol like ``s3://``
        to read from alternative filesystems. Should not include glob
        character(s).
    mode: 'rb', 'wt', etc.
    compression: string
        Compression to use.  See ``dask.bytes.compression.files`` for options.
    encoding: str
        For text mode only
    errors: None or str
        Passed to TextIOWrapper in text mode
    protocol: str or None
        If given, overrides the protocol found in the URL.
    newline: bytes or None
        Used for line terminator in text mode. If None, uses system default;
        if blank, uses no translation.
    **kwargs: dict
        Extra options that make sense to a particular storage connection, e.g.
        host, port, username, password, etc.

    Examples
    --------
    >>> openfile = open('2015-01-01.csv')  # doctest: +SKIP
    >>> openfile = open('s3://bucket/2015-01-01.csv.gz', compression='gzip')  # doctest: +SKIP
    ... with openfile as f:
    ...     df = pd.read_csv(f)

    Returns
    -------
    ``OpenFile`` object.
    """
    return open_files([urlpath], mode, compression, encoding, errors,
                      protocol, newline=newline, **kwargs)[0]


def get_compression(urlpath, compression):
    if compression == 'infer':
        compression = infer_compression(urlpath)
    if compression is not None and compression not in compr:
        raise ValueError("Compression type %s not supported" % compression)
    return compression


def split_protocol(urlpath):
    """Return protocol, path pair"""
    urlpath = stringify_path(urlpath)
    if "://" in urlpath:
        protocol, path = urlpath.split("://", 1)
        if len(protocol) > 1:
            # excludes Windows paths
            return protocol, path
    return None, urlpath


def strip_protocol(urlpath):
    """Return only path part of full URL, according to appropriate backend"""
    protocol, _ = split_protocol(urlpath)
    cls = get_filesystem_class(protocol)
    return cls._strip_protocol(urlpath)


def expand_paths_if_needed(paths, mode, num, fs, name_function):
    """Expand paths if they have a ``*`` in them.

    :param paths: list of paths
    mode: str
        Mode in which to open files.
    num: int
        If opening in writing mode, number of files we expect to create.
    fs: filesystem object
    name_function: callable
        If opening in writing mode, this callable is used to generate path
        names. Names are generated for each partition by
        ``urlpath.replace('*', name_function(partition_index))``.
    :return: list of paths
    """
    expanded_paths = []
    paths = list(paths)
    if 'w' in mode and sum([1 for p in paths if '*' in p]) > 1:
        raise ValueError("When writing data, only one filename mask can "
                         "be specified.")
    elif 'w' in mode:
        num = max(num, len(paths))
    for curr_path in paths:
        if '*' in curr_path:
            if 'w' in mode:
                # expand using name_function
                expanded_paths.extend(
                    _expand_paths(curr_path, name_function, num))
            else:
                # expand using glob
                expanded_paths.extend(fs.glob(curr_path))
        else:
            expanded_paths.append(curr_path)
    # if we generated more paths that asked for, trim the list
    if 'w' in mode and len(expanded_paths) > num:
        expanded_paths = expanded_paths[:num]
    return expanded_paths


def get_fs_token_paths(urlpath, mode='rb', num=1, name_function=None,
                       storage_options=None, protocol=None):
    """Filesystem, deterministic token, and paths from a urlpath and options.

    Parameters
    ----------
    urlpath: string or iterable
        Absolute or relative filepath, URL (may include protocols like
        ``s3://``), or globstring pointing to data.
    mode: str, optional
        Mode in which to open files.
    num: int, optional
        If opening in writing mode, number of files we expect to create.
    name_function: callable, optional
        If opening in writing mode, this callable is used to generate path
        names. Names are generated for each partition by
        ``urlpath.replace('*', name_function(partition_index))``.
    storage_options: dict, optional
        Additional keywords to pass to the filesystem class.
    protocol: str or None
        To override the protocol specifier in the URL
    """
    if isinstance(urlpath, (list, tuple)):
        if not urlpath:
            raise ValueError("empty urlpath sequence")
        protocols, paths = zip(*map(split_protocol, urlpath))
        protocol = protocol or protocols[0]
        if not all(p == protocol for p in protocols):
            raise ValueError("When specifying a list of paths, all paths must "
                             "share the same protocol")
        cls = get_filesystem_class(protocol)
        optionss = list(map(cls._get_kwargs_from_urls, urlpath))
        paths = [cls._strip_protocol(u) for u in urlpath]
        options = optionss[0]
        if not all(o == options for o in optionss):
            raise ValueError("When specifying a list of paths, all paths must "
                             "share the same file-system options")
        update_storage_options(options, storage_options)
        fs = cls(**options)
        paths = expand_paths_if_needed(paths, mode, num, fs, name_function)

    elif isinstance(urlpath, str) or hasattr(urlpath, 'name'):
        protocols, path = split_protocol(urlpath)
        protocol = protocol or protocols
        cls = get_filesystem_class(protocol)

        options = cls._get_kwargs_from_urls(urlpath)
        path = cls._strip_protocol(urlpath)
        update_storage_options(options, storage_options)
        fs = cls(**options)

        if 'w' in mode:
            paths = _expand_paths(path, name_function, num)
        elif "*" in path:
            paths = sorted(fs.glob(path))
        else:
            paths = [path]

    else:
        raise TypeError('url type not understood: %s' % urlpath)

    return fs, fs._fs_token, paths


def _expand_paths(path, name_function, num):
    if isinstance(path, str):
        if path.count('*') > 1:
            raise ValueError("Output path spec must contain exactly one '*'.")
        elif "*" not in path:
            path = os.path.join(path, "*.part")

        if name_function is None:
            name_function = build_name_function(num - 1)

        paths = [path.replace('*', name_function(i)) for i in range(num)]
        if paths != sorted(paths):
            logger.warning("In order to preserve order between partitions"
                           " paths created with ``name_function`` should "
                           "sort to partition order")
    elif isinstance(path, (tuple, list)):
        assert len(path) == num
        paths = list(path)
    else:
        raise ValueError("Path should be either\n"
                         "1. A list of paths: ['foo.json', 'bar.json', ...]\n"
                         "2. A directory: 'foo/\n"
                         "3. A path with a '*' in it: 'foo.*.json'")
    return paths


class BaseCache(object):
    """Pass-though cache: doesn't keep anything, calls every time

    Acts as base class for other cachers

    Parameters
    ----------
    blocksize: int
        How far to read ahead in numbers of bytes
    fetcher: func
        Function of the form f(start, end) which gets bytes from remote as
        specified
    size: int
        How big this file is
    """
    def __init__(self, blocksize, fetcher, size, **kwargs):
        self.blocksize = blocksize
        self.fetcher = fetcher
        self.size = size

    def _fetch(self, start, end):
        return self.fetcher(start, end)


class MMapCache(BaseCache):
    """memory-mapped sparse file cache

    Opens temporary file, which is filled blocks-wise when data is requested.
    Ensure there is enough disc space in the temporary location.

    This cache method might only work on posix
    """

    def __init__(self, blocksize, fetcher, size, location=None,
                 blocks=None, **kwargs):
        super().__init__(blocksize, fetcher, size)
        self.blocks = set() if blocks is None else blocks
        self.location = location
        self.cache = self._makefile()

    def _makefile(self):
        import tempfile
        import mmap

        if self.size == 0:
            return bytearray()

        # posix version
        if self.location is None or not os.path.exists(self.location):
            if self.location is None:
                fd = tempfile.TemporaryFile()
                self.blocks = set()
            else:
                fd = io.open(self.location, 'wb+')
            fd.seek(self.size - 1)
            fd.write(b'1')
            fd.flush()
        else:
            fd = io.open(self.location, 'rb+')

        return mmap.mmap(fd.fileno(), self.size)

    def _fetch(self, start, end):
        start_block = start // self.blocksize
        end_block = end // self.blocksize
        need = [i for i in range(start_block, end_block + 1)
                if i not in self.blocks]
        while need:
            # TODO: not a for loop so we can consolidate blocks later to
            # make fewer fetch calls; this could be parallel
            i = need.pop(0)
            sstart = i * self.blocksize
            send = min(sstart + self.blocksize, self.size)
            self.cache[sstart:send] = self.fetcher(sstart, send)
            self.blocks.add(i)

        return self.cache[start:end]

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state['cache']
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.__dict__.update(state)
        self.cache = self._makefile()


class ReadAheadCache(BaseCache):
    """ Cache which reads only when we get beyond a block of data

    This is a much simpler version of BytesCache, and does not attempt to
    fill holes in the cache or keep fragments alive. It is besst suited to
    many small reads in a sequential order (e.g., readling lines from a file).
    """

    def __init__(self, blocksize, fetcher, size, **kwargs):
        super().__init__(blocksize, fetcher, size)
        self.cache = b''
        self.start = 0
        self.end = 0

    def _fetch(self, start, end):
        end = min(self.size, end)
        l = end - start
        if start >= self.size:
            return b''
        elif start >= self.start and end <= self.end:
            # cache hit
            return self.cache[start - self.start:end - self.start]
        elif self.start <= start < self.end:
            # partial hit
            part = self.cache[start - self.start:]
            l -= len(part)
            start = self.end
        else:
            # miss
            part = b''
        end = min(self.size, end + self.blocksize)
        self.cache = self.fetcher(start, end)  # new block replaces old
        self.start = start
        self.end = self.start + len(self.cache)
        return part + self.cache[:l]


class BytesCache(BaseCache):
    """Cache which holds data in a in-memory bytes object

    Implements read-ahead by the block size, for semi-random reads progressing
    through the file.

    Parameters
    ----------
    trim: bool
        As we read more data, whether to discard the start of the buffer when
        we are more than a blocksize ahead of it.
    """

    def __init__(self, blocksize, fetcher, size, trim=True, **kwargs):
        super().__init__(blocksize, fetcher, size)
        self.cache = b''
        self.start = None
        self.end = None
        self.trim = trim

    def _fetch(self, start, end):
        # TODO: only set start/end after fetch, in case it fails?
        # is this where retry logic might go?
        if (self.start is not None and start >= self.start and self.end is not
                None and end < self.end):
            # cache hit: we have all the required data
            offset = start - self.start
            return self.cache[offset:offset + end - start]
        if self.blocksize:
            bend = min(self.size, end + self.blocksize)
        else:
            bend = end
        if bend == start or start > self.size:
            return b""
        if ((self.start is None or start < self.start) and
                (self.end is None or end > self.end)):
            # First read, or extending both before and after
            self.cache = self.fetcher(start, bend)
            self.start = start
        elif start < self.start:
            if self.end - end > self.blocksize:
                self.cache = self.fetcher(start, bend)
                self.start = start
            else:
                new = self.fetcher(start, self.start)
                self.start = start
                self.cache = new + self.cache
        elif bend > self.end:
            if self.end > self.size:
                pass
            elif end - self.end > self.blocksize:
                self.cache = self.fetcher(start, bend)
                self.start = start
            else:
                new = self.fetcher(self.end, bend)
                self.cache = self.cache + new

        self.end = self.start + len(self.cache)
        offset = start - self.start
        out = self.cache[offset:offset + end - start]
        if self.trim:
            num = (self.end - self.start) // (self.blocksize + 1)
            if num > 1:
                self.start += self.blocksize * num
                self.cache = self.cache[self.blocksize * num:]
        return out

    def __len__(self):
        return len(self.cache)


caches = {'none': BaseCache,
          'mmap': MMapCache,
          'bytes': BytesCache,
          'readahead': ReadAheadCache}



import os
import pytest
from fsspec.implementations.memory import MemoryFileSystem
pytest.importorskip("fuse")
from fsspec.fuse import run
import time


def test_basic(tmpdir):
    tmpdir = str(tmpdir)
    fs = MemoryFileSystem()
    fs.touch('/mounted/testfile')
    th = run(fs, '/mounted/', tmpdir, False)
    timeout = 10
    while True:
        try:
            # can fail with device not ready while waiting for fuse
            if 'testfile' in os.listdir(tmpdir):
                break
        except:
            pass
        timeout -= 1
        time.sleep(1)
        assert timeout > 0, "Timeout"

    fn = os.path.join(tmpdir, 'test')
    with open(fn, 'wb') as f:
        f.write(b'data')
    assert fs.info("/mounted/test")['size'] == 4

    assert open(fn).read() == "data"
    os.remove(fn)

    os.mkdir(fn)
    assert os.listdir(fn) == []

    os.mkdir(fn + '/inner')

    with pytest.raises(OSError):
        os.rmdir(fn)

    os.rmdir(fn + '/inner')
    os.rmdir(fn)
    assert not fs.pseudo_dirs

    # should not normally kill a thread like this, but FUSE blocks, so we
    # cannot have thread listen for event. Alternative may be to .join() but
    # send a SIGINT
    th._tstate_lock.release()
    th._stop()
    th.join()
    fs.store.clear()



import inspect
from distributed.worker import get_worker
from distributed.client import _get_global_client
import dask
from fsspec.spec import AbstractFileSystem, AbstractBufferedFile
from fsspec import filesystem

methods_to_forward = ['mkdir', 'mkdirs', 'rmdir', 'ls', 'copy', 'mv', 'rm']
_files_cache = {}


class DaskWorkerFileSystem(AbstractFileSystem):
    """View files accessible to a worker as any other remote file-system

    When instances are run on the worker, uses the real filesystem. When
    run on the client, they call the worker to provide information or data.

    **Warning** this implementation is experimental, and read-only for now.
    """

    def __init__(self, remote_protocol, storage_options=None, **kwargs):
        super().__init__(**kwargs)
        self.protocol = remote_protocol
        self.storage_options = storage_options or {}
        self.worker = None
        self.client = None
        self.fs = None
        self._determine_worker()

    def _determine_worker(self):
        try:
            get_worker()
            self.worker = True
            self.fs = filesystem(self.protocol, **self.storage_options)
        except ValueError:
            self.worker = False
            self.client = _get_global_client()
            self.rfs = dask.delayed(self)

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._determine_worker()

    def __getstate__(self):
        dic = self.__dict__.copy()
        for method in methods_to_forward + ['client', 'fs']:
            dic.pop(method, None)
        return dic

    def mkdir(self, *args, **kwargs):
        if self.worker:
            self.fs.mkdir(*args, **kwargs)
        else:
            self.rfs.mkdir(*args, **kwargs).compute()

    def rm(self, *args, **kwargs):
        if self.worker:
            self.fs.rm(*args, **kwargs)
        else:
            self.rfs.rm(*args, **kwargs).compute()

    def copy(self, *args, **kwargs):
        if self.worker:
            self.fs.copy(*args, **kwargs)
        else:
            self.rfs.copy(*args, **kwargs).compute()

    def mv(self, *args, **kwargs):
        if self.worker:
            self.fs.mv(*args, **kwargs)
        else:
            self.rfs.mv(*args, **kwargs).compute()

    def ls(self, *args, **kwargs):
        if self.worker:
            return self.fs.ls(*args, **kwargs)
        else:
            return self.rfs.ls(*args, **kwargs).compute()

    def _open(self, path, mode='rb', **kwargs):
        if self.worker:
            return self.fs._open(path, mode=mode)
        else:
            return DaskFile(self, path, mode, **kwargs)

    def fetch_range(self, path, mode, start, end):
        if self.worker:
            with self._open(path, mode) as f:
                f.seek(start)
                return f.read(end-start)
        else:
            return self.rfs.fetch_range(path, mode, start, end).compute()


class DaskFile(AbstractBufferedFile):

    def __init__(self, fs, path, mode='rb', block_size='default',
                 autocommit=True, cache_type='bytes', **kwargs):
        super().__init__(fs, path, mode=mode, block_size=block_size,
                         autocommit=autocommit, cache_type=cache_type,
                         **kwargs)

    def _upload_chunk(self, final=False):
        pass

    def _initiate_upload(self):
        """ Create remote file/upload """
        pass

    def _fetch_range(self, start, end):
        """Get the specified set of bytes from remote"""
        return self.fs.fetch_range(self.path, self.mode, start, end)




from collections.abc import MutableMapping
from .registry import get_filesystem_class
from .core import split_protocol


class FSMap(MutableMapping):
    """Wrap a FileSystem instance as a mutable wrapping.

    The keys of the mapping become files under the given root, and the
    values (which must be bytes) the contents of those files.

    Parameters
    ----------
    root: string
        prefix for all the files
    fs: FileSystem instance
    check: bool (=True)
        performs a touch at the location, to check for write access.

    Examples
    --------
    >>> fs = FileSystem(**parameters) # doctest: +SKIP
    >>> d = FSMap('my-data/path/', fs) # doctest: +SKIP
    or, more likely
    >>> d = fs.get_mapper('my-data/path/')

    >>> d['loc1'] = b'Hello World' # doctest: +SKIP
    >>> list(d.keys()) # doctest: +SKIP
    ['loc1']
    >>> d['loc1'] # doctest: +SKIP
    b'Hello World'
    """

    def __init__(self, root, fs, check=False, create=False):
        self.fs = fs
        self.root = fs._strip_protocol(root).rstrip('/')  # we join on '/' in _key_to_str
        if create:
            if not self.fs.exists(root):
                self.fs.mkdir(root)
        if check:
            if not self.fs.exists(root):
                raise ValueError("Path %s does not exist. Create "
                                 " with the ``create=True`` keyword" %
                                 root)
            self.fs.touch(root+'/a')
            self.fs.rm(root+'/a')

    def clear(self):
        """Remove all keys below root - empties out mapping
        """
        try:
            self.fs.rm(self.root, True)
            self.fs.mkdir(self.root)
        except:
            pass

    def _key_to_str(self, key):
        """Generate full path for the key"""
        if isinstance(key, (tuple, list)):
            key = str(tuple(key))
        else:
            key = str(key)
        return '/'.join([self.root, key]) if self.root else key

    def _str_to_key(self, s):
        """Strip path of to leave key name"""
        return s[len(self.root):].lstrip('/')

    def __getitem__(self, key, default=None):
        """Retrieve data"""
        key = self._key_to_str(key)
        try:
            result = self.fs.cat(key)
        except:
            if default is not None:
                return default
            raise KeyError(key)
        return result

    def pop(self, key, default=None):
        result = self.__getitem__(key, default)
        try:
            del self[key]
        except KeyError:
            pass
        return result

    def __setitem__(self, key, value):
        """Store value in key"""
        key = self._key_to_str(key)
        self.fs.mkdirs(self.fs._parent(key), exist_ok=True)
        with self.fs.open(key, 'wb') as f:
            f.write(value)

    def keys(self):
        """List currently defined keys"""
        return (self._str_to_key(x)
                for x in self.fs.find(self.root))

    def __iter__(self):
        return self.keys()

    def __len__(self):
        return len(self.keys())

    def __delitem__(self, key):
        """Remove key"""
        try:
            self.fs.rm(self._key_to_str(key))
        except:
            raise KeyError

    def __contains__(self, key):
        """Does key exist in mapping?"""
        return self.fs.exists(self._key_to_str(key))

    def __len__(self):
        """Number of stored elements"""
        return sum(1 for _ in self.keys())

    def __getstate__(self):
        """Mapping should be pickleable"""
        # TODO: replace with reduce to reinstantiate?
        return self.fs, self.root

    def __setstate__(self, state):
        fs, root = state
        self.fs = fs
        self.root = root


def get_mapper(url, check=False, create=False, **kwargs):
    """Create key-value interface for given URL and options

    The URL will be of the form "protocol://location" and point to the root
    of the mapper required. All keys will be file-names below this location,
    and their values the contents of each key.

    Parameters
    ----------
    url: str
        Root URL of mapping
    check: bool
        Whether to attempt to read from the location before instantiation, to
        check that the mapping does exist
    create: bool
        Whether to make the directory corresponding to the root before
        instantiating

    Returns
    -------
    ``FSMap`` instance, the dict-like key-value store.
    """
    protocol, path = split_protocol(url)
    cls = get_filesystem_class(protocol)
    fs = cls(**kwargs)
    # Removing protocol here - could defer to each open() on the backend
    return FSMap(path, fs, check, create)



import pickle
import pytest
import subprocess
import time

requests = pytest.importorskip("requests")
from fsspec.implementations.webhdfs import WebHDFS
import fsspec


@pytest.fixture(scope='module')
def hdfs_cluster():
    cmd0 = "htcluster shutdown".split()
    try:
        subprocess.check_output(cmd0)
    except FileNotFoundError:
        pytest.skip("htcluster not found")
    cmd1 = "htcluster startup --image base".split()
    subprocess.check_output(cmd1)
    try:
        while True:
            t = 90
            try:
                requests.get('http://localhost:50070/webhdfs/v1/?op=LISTSTATUS')
            except:
                t -= 1
                assert t > 0, "Timeout waiting for HDFS"
                time.sleep(1)
                continue
            break
        time.sleep(7)
        yield "localhost"
    finally:
        subprocess.check_output(cmd0)


def test_pickle(hdfs_cluster):
    w = WebHDFS(hdfs_cluster, user='testuser')
    w2 = pickle.loads(pickle.dumps(w))
    assert w == w2


def test_simple(hdfs_cluster):
    w = WebHDFS(hdfs_cluster, user='testuser')
    home = w.home_directory()
    assert home == '/user/testuser'
    with pytest.raises(PermissionError):
        w.mkdir('/root')


def test_url(hdfs_cluster):
    url = 'webhdfs://testuser@localhost:50070/user/testuser/myfile'
    fo = fsspec.open(url, 'wb', data_proxy={'worker.example.com': 'localhost'})
    with fo as f:
        f.write(b'hello')
    fo = fsspec.open(url, 'rb', data_proxy={'worker.example.com': 'localhost'})
    with fo as f:
        assert f.read() == b'hello'


def test_workflow(hdfs_cluster):
    w = WebHDFS(hdfs_cluster, user='testuser',
                data_proxy={'worker.example.com': 'localhost'})
    fn = '/user/testuser/testrun/afile'
    w.mkdir('/user/testuser/testrun')
    with w.open(fn, 'wb') as f:
        f.write(b'hello')
    assert w.exists(fn)
    info = w.info(fn)
    assert info['size'] == 5
    assert w.isfile(fn)
    assert w.cat(fn) == b'hello'
    w.rm('/user/testuser/testrun', recursive=True)
    assert not w.exists(fn)


def test_with_gzip(hdfs_cluster):
    from gzip import GzipFile
    w = WebHDFS(hdfs_cluster, user='testuser',
                data_proxy={'worker.example.com': 'localhost'})
    fn = '/user/testuser/gzfile'
    with w.open(fn, 'wb') as f:
        gf = GzipFile(fileobj=f, mode='w')
        gf.write(b'hello')
        gf.close()
    with w.open(fn, 'rb') as f:
        gf = GzipFile(fileobj=f, mode='r')
        assert gf.read() == b'hello'


def test_workflow_transaction(hdfs_cluster):
    w = WebHDFS(hdfs_cluster, user='testuser',
                data_proxy={'worker.example.com': 'localhost'})
    fn = '/user/testuser/testrun/afile'
    w.mkdirs('/user/testuser/testrun')
    with w.transaction:
        with w.open(fn, 'wb') as f:
            f.write(b'hello')
        assert not w.exists(fn)
    assert w.exists(fn)
    assert w.ukey(fn)
    files = w.ls('/user/testuser/testrun', True)
    summ = w.content_summary('/user/testuser/testrun')
    assert summ['length'] == files[0]['size']
    assert summ['fileCount'] == 1

    w.rm('/user/testuser/testrun', recursive=True)
    assert not w.exists(fn)




import time
import pickle
import logging
import os
import hashlib
import tempfile
import inspect
from fsspec import AbstractFileSystem, filesystem
from fsspec.core import MMapCache
logger = logging.getLogger('fsspec')


class CachingFileSystem(AbstractFileSystem):
    """Locally caching filesystem, layer over any other FS

    This class implements chunk-wise local storage of remote files, for quick
    access after the initial download. The files are stored in a given
    directory with random hashes for the filenames. If no directory is given,
    a temporary one is used, which should be cleaned up by the OS after the
    process ends. The files themselves as sparse (as implemented in
    MMapCache), so only the data which is accessed takes up space.

    Restrictions:

    - the block-size must be the same for each access of a given file, unless
      all blocks of the file have already been read
    - caching can only be applied to file-systems which produce files
      derived from fsspec.spec.AbstractBufferedFile ; LocalFileSystem is also
      allowed, for testing
    """

    protocol = ('blockcache', 'cached')

    def __init__(self, target_protocol=None, cache_storage='TMP',
                 cache_check=10, check_files=False,
                 expiry_time=604800, target_options=None):
        """

        Parameters
        ----------
        target_protocol: str
            Target fielsystem protocol
        cache_storage: str or list(str)
            Location to store files. If "TMP", this is a temporary directory,
            and will be cleaned up by the OS when this process ends (or later).
            If a list, each location will be tried in the order given, but
            only the last will be considered writable.
        cache_check: int
            Number of seconds between reload of cache metadata
        check_files: bool
            Whether to explicitly see if the UID of the remote file matches
            the stored one before using. Warning: some file systems such as
            HTTP cannot reliably give a unique hash of the contents of some
            path, so be sure to set this option to False.
        expiry_time: int
            The time in seconds after which a local copy is considered useless.
            Set to falsy to prevent expiry. The default is equivalent to one
            week.
        target_options: dict or None
            Passed to the instantiation of the FS, if fs is None.
        """
        if cache_storage == "TMP":
            storage = [tempfile.mkdtemp()]
        else:
            if isinstance(cache_storage, str):
                storage = [cache_storage]
            else:
                storage = cache_storage
        os.makedirs(storage[-1], exist_ok=True)
        self.protocol = target_protocol
        self.storage = storage
        self.kwargs = target_options or {}
        self.cache_check = cache_check
        self.check_files = check_files
        self.expiry = expiry_time
        self.load_cache()
        self.fs = filesystem(target_protocol, **self.kwargs)
        super().__init__(**self.kwargs)

    def __reduce_ex__(self, *_):
        return self.__class__, (
            self.protocol, self.storage, self.cache_check, self.check_files,
            self.expiry, self.kwargs or None)

    def load_cache(self):
        """Read set of stored blocks from file"""
        cached_files = []
        for storage in self.storage:
            fn = os.path.join(storage, 'cache')
            if os.path.exists(fn):
                with open(fn, 'rb') as f:
                    # TODO: consolidate blocks here
                    cached_files.append(pickle.load(f))
            else:
                os.makedirs(storage, exist_ok=True)
                cached_files.append({})
        self.cached_files = cached_files or [{}]
        self.last_cache = time.time()

    def save_cache(self):
        """Save set of stored blocks from file"""
        fn = os.path.join(self.storage[-1], 'cache')
        # TODO: a file lock could be used to ensure file does not change
        #  between re-read and write; but occasional duplicated reads ok.
        cache = self.cached_files[-1]
        if os.path.exists(fn):
            with open(fn, 'rb') as f:
                cached_files = pickle.load(f)
            for k, c in cached_files.items():
                if c['blocks'] is not True:
                    if cache[k]['blocks'] is True:
                        c['blocks'] = True
                    else:
                        c['blocks'] = c['blocks'].union(cache[k]['blocks'])
        else:
            cached_files = cache
        cache = {k: v.copy() for k, v in cached_files.items()}
        for c in cache.values():
            if isinstance(c['blocks'], set):
                c['blocks'] = list(c['blocks'])
        with open(fn + '.temp', 'wb') as f:
            pickle.dump(cache, f)
        if os.path.exists(fn):
            os.remove(fn)
        os.rename(fn + '.temp', fn)

    def _check_cache(self):
        """Reload caches if time elapsed or any disappeared"""
        if not self.cache_check:
            # explicitly told not to bother checking
            return
        timecond = time.time() - self.last_cache > self.cache_check
        existcond = all(os.path.exists(storage) for storage in self.storage)
        if timecond or not existcond:
            self.load_cache()

    def _check_file(self, path):
        """Is path in cache and still valid"""
        self._check_cache()
        for storage, cache in zip(self.storage, self.cached_files):
            if path not in cache:
                continue
            detail = cache[path].copy()
            if self.check_files:
                if detail['uid'] != self.fs.ukey(path):
                    continue
            if self.expiry:
                if detail['time'] - time.time() > self.expiry:
                    continue
            fn = os.path.join(storage, detail['fn'])
            if os.path.exists(fn):
                return detail, fn
        return False, None

    def _open(self, path, mode='rb', **kwargs):
        """Wrap the target _open

        If the whole file exists in the cache, just open it locally and
        return that.

        Otherwise, open the file on the target FS, and make it have a mmap
        cache pointing to the location which we determine, in our cache.
        The ``blocks`` instance is shared, so as the mmap cache instance
        updates, so does the entry in our ``cached_files`` attribute.
        We monkey-patch this file, so that when it closes, we call
        ``close_and_update`` to save the state of the blocks.
        """
        path = self._strip_protocol(path)
        if not path.startswith(self.protocol):
            path = self.protocol + "://" + path
        if mode != 'rb':
            return self.fs._open(path, mode=mode, **kwargs)
        detail, fn = self._check_file(path)
        if detail:
            # file is in cache
            hash, blocks = detail['fn'], detail['blocks']
            if blocks is True:
                # stored file is complete
                logger.debug("Opening local copy of %s" % path)
                return open(fn, 'rb')
            # TODO: action where partial file exists in read-only cache
            logger.debug("Opening partially cached copy of %s" % path)
        else:
            hash = hashlib.sha256(path.encode()).hexdigest()
            fn = os.path.join(self.storage[-1], hash)
            blocks = set()
            detail = {'fn': hash, 'blocks': blocks,
                      'time': time.time(),
                      'uid': self.fs.ukey(path)}
            self.cached_files[-1][path] = detail
            logger.debug("Creating local sparse file for %s" % path)
        kwargs['cache_type'] = 'none'
        kwargs['mode'] = mode

        # call target filesystems open
        f = self.fs._open(path, **kwargs)
        if 'blocksize' in detail:
            if detail['blocksize'] != f.blocksize:
                raise ValueError('Cached file must be reopened with same block'
                                 'size as original (old: %i, new %i)'
                                 '' % (detail['blocksize'], f.blocksize))
        else:
            detail['blocksize'] = f.blocksize
        f.cache = MMapCache(f.blocksize, f._fetch_range, f.size,
                            fn, blocks)
        close = f.close
        f.close = lambda: self.close_and_update(f, close)
        return f

    def close_and_update(self, f, close):
        """Called when a file is closing, so store the set of blocks"""
        if f.path.startswith(self.protocol):
            path = f.path
        else:
            path = self.protocol + "://" + f.path
        c = self.cached_files[-1][path]
        if (c['blocks'] is not True
                and len(['blocks']) * f.blocksize >= f.size):
            c['blocks'] = True
        self.save_cache()
        close()

    def __getattribute__(self, item):
        if item in ['load_cache', '_open', 'save_cache', 'close_and_update',
                    '__init__', '__getattribute__', '__reduce_ex__', 'open',
                    'cat', 'get', 'read_block', 'tail', 'head',
                    '_check_file', '_check_cache']:
            # all the methods defined in this class. Note `open` here, since
            # it calls `_open`, but is actually in superclass
            return lambda *args, **kw: getattr(type(self), item)(
                self, *args, **kw
            )
        if item == '__class__':
            return type(self)
        d = object.__getattribute__(self, '__dict__')
        fs = d.get('fs', None)  # fs is not immediately defined
        if item in d:
            return d[item]
        elif fs is not None:
            if item in fs.__dict__:
                # attribute of instance
                return fs.__dict__[item]
            # attributed belonging to the target filesystem
            cls = type(fs)
            m = getattr(cls, item)
            if (inspect.isfunction(m) and (not hasattr(m, '__self__')
                                           or m.__self__ is None)):
                # instance method
                return m.__get__(fs, cls)
            return m  # class method or attribute
        else:
            # attributes of the superclass, while target is being set up
            return super().__getattribute__(item)


class WholeFileCacheFileSystem(CachingFileSystem):
    """Caches whole remote files on first access

    This class is intended as a layer over any other file system, and
    will make a local copy of each file accessed, so that all subsequent
    reads are local. This is similar to ``CachingFileSystem``, but without
    the block-wise functionality and so can work even when sparse files
    are not allowed. See its docstring for definition of the init
    arguments.

    The class still needs access to the remote store for listing files,
    and may refresh cached files.
    """
    protocol = 'filecache'

    def _open(self, path, mode='rb', **kwargs):
        path = self._strip_protocol(path)
        if not path.startswith(self.protocol):
            path = self.protocol + "://" + path
        if mode != 'rb':
            return self.fs._open(path, mode=mode, **kwargs)
        detail, fn = self._check_file(path)
        if detail:
            hash, blocks = detail['fn'], detail['blocks']
            if blocks is True:
                logger.debug("Opening local copy of %s" % path)
                return open(fn, 'rb')
            else:
                raise ValueError("Attempt to open partially cached file %s"
                                 "as a wholly cached file" % path)
        else:
            hash = hashlib.sha256(path.encode()).hexdigest()
            fn = os.path.join(self.storage[-1], hash)
            blocks = True
            detail = {'fn': hash, 'blocks': blocks,
                      'time': time.time(),
                      'uid': self.fs.ukey(path)}
            self.cached_files[-1][path] = detail
            logger.debug("Copying %s to local cache" % path)
        kwargs['cache_type'] = 'none'
        kwargs['mode'] = mode

        # call target filesystems open
        f = self.fs._open(path, **kwargs)
        with open(fn, 'wb') as f2:
            if f.blocksize and f.size:
                # opportunity to parallelise here
                data = True
                while data:
                    data = f.read(f.blocksize)
                    f2.write(data)
            else:
                # this only applies to HTTP, should instead use streaming
                f2.write(f.read())
        self.save_cache()
        return self._open(path, mode)



import glob
import os
import pytest
from http.server import BaseHTTPRequestHandler, HTTPServer
import threading
import fsspec

requests = pytest.importorskip('requests')
port = 9898
data = b'\n'.join([b'some test data'] * 1000)
realfile = "http://localhost:%i/index/realfile" % port
index = b'<a href="%s">Link</a>' % realfile.encode()


class HTTPTestHandler(BaseHTTPRequestHandler):
    def _respond(self, code=200, headers=None, data=b''):
        headers = headers or {}
        headers.update({'User-Agent': 'test'})
        self.send_response(code)
        for k, v in headers.items():
            self.send_header(k, str(v))
        self.end_headers()
        if data:
            self.wfile.write(data)

    def do_GET(self):
        if self.path not in ['/index/realfile', '/index']:
            self._respond(404)
            return

        d = data if self.path == '/index/realfile' else index
        if 'Range' in self.headers:
            ran = self.headers['Range']
            b, ran = ran.split("=")
            start, end = ran.split('-')
            print(start)
            print(end)
            d = d[int(start):int(end)+1]
        if 'give_length' in self.headers:
            self._respond(200, {'Content-Length': len(d)}, d)
        elif 'give_range' in self.headers:
            self._respond(
                200, {'Content-Range': "0-%i/%i" % (len(d) - 1, len(d))},
                d
            )
        else:
            self._respond(200, data=d)

    def do_HEAD(self):
        if 'head_ok' not in self.headers:
            self._respond(405)
            return
        d = data if self.path == '/index/realfile' else index
        if self.path not in ['/index/realfile', '/index']:
            self._respond(404)
        elif 'give_length' in self.headers:
            self._respond(200, {'Content-Length': len(d)})
        elif 'give_range' in self.headers:
            self._respond(200, {'Content-Range': "0-%i/%i" %
                                                 (len(d) - 1, len(d))})
        else:
            self._respond(200)  # OK response, but no useful info


@pytest.fixture(scope='module')
def server():
    server_address = ('', port)
    httpd = HTTPServer(server_address, HTTPTestHandler)
    th = threading.Thread(target=httpd.serve_forever)
    th.daemon = True
    th.start()
    try:
        yield 'http://localhost:%i' % port
    finally:
        httpd.socket.close()
        httpd.shutdown()
        th.join()


def test_list(server):
    h = fsspec.filesystem('http')
    out = h.glob(server + '/index/*')
    assert out == [server + '/index/realfile']


def test_policy_arg(server):
    h = fsspec.filesystem('http', size_policy='get')
    out = h.glob(server + '/index/*')
    assert out == [server + '/index/realfile']


def test_exists(server):
    h = fsspec.filesystem('http')
    assert not h.exists(server + '/notafile')


def test_read(server):
    h = fsspec.filesystem('http')
    out = server + '/index/realfile'
    with h.open(out, 'rb') as f:
        assert f.read() == data
    with h.open(out, 'rb', block_size=0) as f:
        assert f.read() == data
    with h.open(out, 'rb') as f:
        assert f.read(100) + f.read() == data


def test_methods(server):
    h = fsspec.filesystem('http')
    url = server + '/index/realfile'
    assert h.exists(url)
    assert h.cat(url) == data


@pytest.mark.parametrize('headers', [{},
                                     {'give_length': 'true'},
                                     {'give_length': 'true', 'head_ok': 'true'},
                                     {'give_range': 'true'}
                                     ])
def test_random_access(server, headers):
    h = fsspec.filesystem('http', headers=headers)
    url = server + '/index/realfile'
    with h.open(url, 'rb') as f:
        if headers:
            assert f.size == len(data)
        assert f.read(5) == data[:5]
        # python server does not respect bytes range request
        # we actually get all the data
        f.seek(5, 1)
        assert f.read(5) == data[10:15]




import paramiko
from stat import S_ISDIR, S_ISLNK
import types
import uuid
from .. import AbstractFileSystem
from ..utils import infer_storage_options


class SFTPFileSystem(AbstractFileSystem):
    """Files over SFTP/SSH

    Peer-to-peer filesystem over SSH using paramiko.
    """
    protocol = 'sftp', 'ssh'

    def __init__(self, host, **ssh_kwargs):
        """

        Parameters
        ----------
        host: str
            Hostname or IP as a string
        temppath: str
            Location on the server to put files, when within a transaction
        ssh_kwargs: dict
            Parameters passed on to connection. See details in
            http://docs.paramiko.org/en/2.4/api/client.html#paramiko.client.SSHClient.connect
            May include port, username, password...
        """
        super(SFTPFileSystem, self).__init__(**ssh_kwargs)
        self.temppath = ssh_kwargs.pop('temppath', '/tmp')
        self.host = host
        self.ssh_kwargs = ssh_kwargs
        self._connect()

    def _connect(self):
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.connect(self.host, **self.ssh_kwargs)
        self.ftp = self.client.open_sftp()

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('ftp', None)
        d.pop('client', None)
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._connect()

    @classmethod
    def _strip_protocol(cls, path):
        return infer_storage_options(path)['path']

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        out = infer_storage_options(urlpath)
        out.pop('path', None)
        out.pop('protocol', None)
        return out

    def mkdir(self, path, mode=511):
        self.ftp.mkdir(path, mode)

    def makedirs(self, path, mode=511):
        parts = path.split('/')
        path = ''
        for part in parts:
            path += '/' + part
            if not self.exists(path):
                self.mkdir(path, mode)

    def rmdir(self, path):
        self.ftp.rmdir(path)

    def info(self, path):
        s = self.ftp.stat(path)
        if S_ISDIR(s.st_mode):
            t = 'directory'
        elif S_ISLNK(s.st_mode):
            t = 'link'
        else:
            t = 'file'
        return {'name': path + '/' if t == 'directory' else path,
                'size': s.st_size, 'type': t, 'uid': s.st_uid,
                'gui': s.st_gid, 'time': s.st_atime, 'mtime': s.st_mtime}

    def ls(self, path, detail=False):
        out = ['/'.join([path.rstrip('/'), p]) for p in self.ftp.listdir(path)]
        out = [self.info(o) for o in out]
        if detail:
            return out
        return sorted([p['name'] for p in out])

    def put(self, lpath, rpath):
        self.ftp.put(lpath, rpath)

    def get(self, rpath, lpath):
        self.ftp.get(rpath, lpath)

    def _open(self, path, mode='rb', block_size=None, **kwargs):
        """
        block_size: int or None
            If 0, no buffering, if 1, line buffering, if >1, buffer that many
            bytes, if None use default from paramiko.
        """
        if kwargs.get('autocommit', True) is False:
            # writes to temporary file, move on commit
            path2 = "{}/{}".format(self.temppath,uuid.uuid4())
            f = self.ftp.open(path2, mode,
                              bufsize=block_size if block_size else -1)
            f.temppath = path2
            f.targetpath = path
            f.fs = self
            f.commit = types.MethodType(commit_a_file, f)
            f.discard = types.MethodType(discard_a_file, f)
        else:
            f = self.ftp.open(path, mode,
                              bufsize=block_size if block_size else -1)
        return f

    def _rm(self, path):
        if self.isdir(path):
            self.ftp.rmdir(path)
        else:
            self.ftp.remove(path)

    def mv(self, old, new):
        self.ftp.posix_rename(old, new)


def commit_a_file(self):
    self.fs.mv(self.temppath, self.targetpath)


def discard_a_file(self):
    self.fs._rm(self.temppath)



import pytest
import shlex
import subprocess
import time
import fsspec

pytest.importorskip('paramiko')


def stop_docker(name):
    cmd = shlex.split('docker ps -a -q --filter "name=%s"' % name)
    cid = subprocess.check_output(cmd).strip().decode()
    if cid:
        subprocess.call(['docker', 'rm', '-f', cid])


@pytest.fixture(scope='module')
def ssh():
    # requires docker
    cmds = r"""apt-get update
apt-get install -y openssh-server
mkdir /var/run/sshd
bash -c "echo 'root:pass' | chpasswd"
sed -i 's/PermitRootLogin prohibit-password/PermitRootLogin yes/' /etc/ssh/sshd_config
sed 's@session\s*required\s*pam_loginuid.so@session optional pam_loginuid.so@g' -i /etc/pam.d/sshd
bash -c "echo \"export VISIBLE=now\" >> /etc/profile"
/usr/sbin/sshd
""".split('\n')
    name = 'fsspec_sftp'
    stop_docker(name)
    cmd = ('docker run -d -p 9200:22 --name {} '
           'ubuntu:16.04 sleep 9000'.format(name))
    cid = subprocess.check_output(shlex.split(cmd)).strip().decode()
    for cmd in cmds:
        subprocess.call(['docker', 'exec', cid] + shlex.split(cmd))
    try:
        time.sleep(1)
        yield dict(host='localhost', port=9200, username='root',
                   password='pass')
    finally:
        stop_docker(name)


def test_simple(ssh):
    f = fsspec.get_filesystem_class('sftp')(**ssh)
    f.mkdirs('/home/someuser/deeper')
    f.touch('/home/someuser/deeper/afile')
    assert f.find('/home/someuser') == ['/home/someuser/deeper/afile']
    assert f.ls('/home/someuser/deeper/') == ['/home/someuser/deeper/afile']
    assert f.info('/home/someuser/deeper/afile')['type'] == 'file'
    assert f.info('/home/someuser/deeper/afile')['size'] == 0
    assert f.exists('/home/someuser')
    f.rm('/home/someuser', recursive=True)
    assert not f.exists('/home/someuser')


@pytest.mark.parametrize('protocol', ['sftp', 'ssh'])
def test_with_url(protocol, ssh):
    fo = fsspec.open(protocol + "://{username}:{password}@{host}:{port}"
                     "/home/someuserout".format(**ssh), 'wb')
    with fo as f:
        f.write(b'hello')
    fo = fsspec.open(protocol + "://{username}:{password}@{host}:{port}"
                     "/home/someuserout".format(**ssh), 'rb')
    with fo as f:
        assert f.read() == b'hello'



def test_transaction(ssh):
    f = fsspec.get_filesystem_class('sftp')(**ssh)
    f.mkdirs('/home/someuser/deeper')
    f.start_transaction()
    f.touch('/home/someuser/deeper/afile')
    assert f.find('/home/someuser') == []
    f.end_transaction()
    f.find('/home/someuser') == ['/home/someuser/deeper/afile']

    with f.transaction:
        assert f._intrans
        f.touch('/home/someuser/deeper/afile2')
        assert f.find('/home/someuser') == ['/home/someuser/deeper/afile']
    assert f.find('/home/someuser') == ['/home/someuser/deeper/afile',
                                        '/home/someuser/deeper/afile2']





# https://hadoop.apache.org/docs/r1.0.4/webhdfs.html

import requests
from urllib.parse import quote
import uuid
from ..spec import AbstractFileSystem, AbstractBufferedFile
from ..utils import infer_storage_options
import logging
logger = logging.getLogger('webhdfs')


class WebHDFS(AbstractFileSystem):
    """
    Interface to HDFS over HTTP

    Three auth mechanisms are supported:

    insecure: no auth is done, and the user is assumed to be whoever they
        say they are (parameter `user`), or a predefined value such as
        "dr.who" if not given
    spnego: when kerberos authentication is enabled, auth is negotiated by
        requests_kerberos https://github.com/requests/requests-kerberos .
        This establishes a session based on existing kinit login and/or
        specified principal/password; paraneters are passed with ``kerb_kwargs``
    token: uses an existing Hadoop delegation token from another secured
        service. Indeed, this client can also generate such tokens when
        not insecure. Note that tokens expire, but can be renewed (by a
        previously specified user) and may allow for proxying.

    """
    tempdir = '/tmp'
    protocol = 'webhdfs', 'webHDFS'

    def __init__(self, host, port=50070, kerberos=False, token=None, user=None,
                 proxy_to=None, kerb_kwargs=None, data_proxy=None,
                 **kwargs):
        """
        Parameters
        ----------
        host: str
            Name-node address
        port: int
            Port for webHDFS
        kerberos: bool
            Whether to authenticate with kerberos for this connection
        token: str or None
            If given, use this token on every call to authenticate. A user
            and user-proxy may be encoded in the token and should not be also
            given
        user: str or None
            If given, assert the user name to connect with
        proxy_to: str or None
            If given, the user has the authority to proxy, and this value is
            the user in who's name actions are taken
        kerb_kwargs: dict
            Any extra arguments for HTTPKerberosAuth, see
            https://github.com/requests/requests-kerberos/blob/master/requests_kerberos/kerberos_.py
        data_proxy: dict, callable or None
            If given, map data-node addresses. This can be necessary if the
            HDFS cluster is behind a proxy, running on Docker or otherwise has
            a mismatch between the host-names given by the name-node and the
            address by which to refer to them from the client. If a dict,
            maps host names `host->data_proxy[host]`; if a callable, full
            URLs are passed, and function must conform to
            `url->data_proxy(url)`.
        kwargs
        """
        super().__init__(**kwargs)
        self.url = f"http://{host}:{port}/webhdfs/v1"
        self.kerb = kerberos
        self.kerb_kwargs = kerb_kwargs or {}
        self.pars = {}
        self.proxy = data_proxy or {}
        if token is not None:
            if user is not None or proxy_to is not None:
                raise ValueError('If passing a delegation token, must not set '
                                 'user or proxy_to, as these are encoded in the'
                                 ' token')
            self.pars['delegation'] = token
        if user is not None:
            self.pars['user.name'] = user
        if proxy_to is not None:
            self.pars['doas'] = proxy_to
        if kerberos and user is not None:
            raise ValueError('If using Kerberos auth, do not specify the '
                             'user, this is handled by kinit.')
        self._connect()

    def _connect(self):
        self.session = requests.Session()
        if self.kerb:
            from requests_kerberos import HTTPKerberosAuth
            self.session.auth = HTTPKerberosAuth(**self.kerb_kwargs)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('session', None)
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._connect()

    def _call(self, op, method='get', path=None, data=None,
              redirect=True, **kwargs):
        url = self.url + quote(path or "")
        args = kwargs.copy()
        args.update(self.pars)
        args['op'] = op.upper()
        logger.debug(url, method, args)
        out = self.session.request(method=method.upper(), url=url, params=args,
                                   data=data, allow_redirects=redirect)
        if out.status_code == 404:
            raise FileNotFoundError(path)
        if out.status_code == 403:
            raise PermissionError(path or "")
        if out.status_code == 401:
            raise PermissionError  # not specific to path
        out.raise_for_status()
        return out

    def _open(self, path, mode='rb', block_size=None, autocommit=True,
              replication=None, permissions=None, **kwargs):
        """

        Parameters
        ----------
        path: str
            File location
        mode: str
            'rb', 'wb', etc.
        block_size: int
            Client buffer size for read-ahead or write buffer
        autocommit: bool
            If False, writes to temporary file that only gets put in final
            location upon commit
        replication: int
            Number of copies of file on the cluster, write mode only
        permissions: str or int
            posix permissions, write mode only
        kwargs

        Returns
        -------
        WebHDFile instance
        """
        block_size = block_size or self.blocksize
        return WebHDFile(self, path, mode=mode, block_size=block_size,
                         tempdir=self.tempdir, autocommit=autocommit,
                         replication=replication, permissions=permissions)

    @staticmethod
    def _process_info(info):
        info['type'] = info['type'].lower()
        info['size'] = info['length']
        return info

    @classmethod
    def _strip_protocol(cls, path):
        return infer_storage_options(path)['path']

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        out = infer_storage_options(urlpath)
        out.pop('path', None)
        out.pop('protocol', None)
        if 'username' in out:
            out['user'] = out.pop('username')
        return out

    def info(self, path):
        out = self._call('GETFILESTATUS', path=path)
        info = out.json()['FileStatus']
        info['name'] = path
        return self._process_info(info)

    def ls(self, path, detail=False):
        out = self._call('LISTSTATUS', path=path)
        infos = out.json()['FileStatuses']['FileStatus']
        for info in infos:
            self._process_info(info)
            info['name'] = path.rstrip('/') + '/' + info['pathSuffix']
        if detail:
            return sorted(infos, key=lambda i: i['name'])
        else:
            return sorted(info['name'] for info in infos)

    def content_summary(self, path):
        """Total numbers of files, directories and bytes under path"""
        out = self._call('GETCONTENTSUMMARY', path=path)
        return out.json()['ContentSummary']

    def ukey(self, path):
        """Checksum info of file, giving method and result"""
        out = self._call('GETFILECHECKSUM', path=path, redirect=False)
        location = self._apply_proxy(out.headers['Location'])
        out2 = self.session.get(location)
        out2.raise_for_status()
        return out2.json()['FileChecksum']

    def home_directory(self):
        """Get user's home directory"""
        out = self._call('GETHOMEDIRECTORY')
        return out.json()['Path']

    def get_delegation_token(self, renewer=None):
        """Retrieve token which can give the same authority to other uses

        Parameters
        ----------
        renewer: str or None
            User who may use this token; if None, will be current user
        """
        if renewer:
            out = self._call('GETDELEGATIONTOKEN', renewer=renewer)
        else:
            out = self._call('GETDELEGATIONTOKEN')
        t = out.json()["Token"]
        if t is None:
            raise ValueError('No token available for this '
                             'user/security context')
        return t["urlString"]

    def renew_delegation_token(self, token):
        """Make token live longer. Returns new expiry time"""
        out = self._call('RENEWDELEGATIONTOKEN', method='put', token=token)
        return out.json()['long']

    def cancel_delegation_token(self, token):
        """Stop the token from being useful"""
        self._call('CANCELDELEGATIONTOKEN', method='put', token=token)

    def chmod(self, path, mod):
        """Set the permission at path

        Parameters
        ----------
        path: str
            location to set (file or directory)
        mod: str or int
            posix epresentation or permission, give as oct string, e.g, '777'
            or 0o777
        """
        self._call('SETPERMISSION', method='put', path=path, permission=mod)

    def chown(self, path, owner=None, group=None):
        """Change owning user and/or group"""
        kwargs = {}
        if owner is not None:
            kwargs['owner'] = owner
        if group is not None:
            kwargs['group'] = group
        self._call('SETOWNER', method='put', path=path, **kwargs)

    def set_replication(self, path, replication):
        """
        Set file replication factor

        Parameters
        ----------
        path: str
            File location (not for directories)
        replication: int
            Number of copies of file on the cluster. Should be smaller than
            number of data nodes; normally 3 on most systems.
        """
        self._call('SETREPLICATION', path=path, method='put',
                   replication=replication)

    def mkdir(self, path, **kwargs):
        self._call('MKDIRS', method='put', path=path)

    def makedirs(self, path, exist_ok=False):
        if exist_ok is False and self.exists(path):
            raise FileExistsError(path)
        self.mkdir(path)

    def mv(self, path1, path2, **kwargs):
        self._call('RENAME', method='put', path=path1, destination=path2)

    def rm(self, path, recursive=False, **kwargs):
        self._call('DELETE', method='delete', path=path,
                   recursive='true' if recursive else 'false')

    def _apply_proxy(self, location):
        if self.proxy and callable(self.proxy):
            location = self.proxy(location)
        elif self.proxy:
            # as a dict
            for k, v in self.proxy.items():
                location = location.replace(k, v, 1)
        return location


class WebHDFile(AbstractBufferedFile):
    """A file living in HDFS over webHDFS"""

    def __init__(self, fs, path, **kwargs):
        super().__init__(fs, path, **kwargs)
        kwargs = kwargs.copy()
        if kwargs.get('permissions', None) is None:
            kwargs.pop('permissions', None)
        if kwargs.get('replication', None) is None:
            kwargs.pop('replication', None)
        self.permissions = kwargs.pop('permissions', 511)
        tempdir = kwargs.pop('tempdir')
        if kwargs.pop('autocommit', False) is False:
            self.target = self.path
            self.path = '/'.join([tempdir, str(uuid.uuid4())])

    def _upload_chunk(self, final=False):
        """ Write one part of a multi-block file upload

        Parameters
        ==========
        final: bool
            This is the last block, so should complete file, if
            self.autocommit is True.
        """
        out = self.fs.session.post(self.location, data=self.buffer.getvalue())
        out.raise_for_status()
        return True

    def _initiate_upload(self):
        """ Create remote file/upload """
        if 'a' in self.mode:
            op, method = 'APPEND', 'POST'
        else:
            op, method = 'CREATE', 'PUT'
            if self.fs.exists(self.path):
                # no "truncate" or "create empty"
                self.fs.rm(self.path)
        out = self.fs._call(op, method, self.path, redirect=False,
                            **self.kwargs)
        location = self.fs._apply_proxy(out.headers['Location'])
        if 'w' in self.mode:
            # create empty file to append to
            out2 = self.fs.session.put(location)
            out2.raise_for_status()
        self.location = location.replace('CREATE', 'APPEND')

    def _fetch_range(self, start, end):
        out = self.fs._call('OPEN', path=self.path, offset=start,
                            length=end-start, redirect=False)
        out.raise_for_status()
        location = out.headers['Location']
        out2 = self.fs.session.get(self.fs._apply_proxy(location))
        return out2.content

    def commit(self):
        self.fs.mv(self.path, self.target)

    def discard(self):
        self.fs.rm(self.path)



from hashlib import md5
import math
import os
import pathlib
import re
from urllib.parse import urlsplit


DEFAULT_BLOCK_SIZE = 5 * 2 ** 20


def infer_storage_options(urlpath, inherit_storage_options=None):
    """ Infer storage options from URL path and merge it with existing storage
    options.

    Parameters
    ----------
    urlpath: str or unicode
        Either local absolute file path or URL (hdfs://namenode:8020/file.csv)
    inherit_storage_options: dict (optional)
        Its contents will get merged with the inferred information from the
        given path

    Returns
    -------
    Storage options dict.

    Examples
    --------
    >>> infer_storage_options('/mnt/datasets/test.csv')  # doctest: +SKIP
    {"protocol": "file", "path", "/mnt/datasets/test.csv"}
    >>> infer_storage_options(
    ...          'hdfs://username:pwd@node:123/mnt/datasets/test.csv?q=1',
    ...          inherit_storage_options={'extra': 'value'})  # doctest: +SKIP
    {"protocol": "hdfs", "username": "username", "password": "pwd",
    "host": "node", "port": 123, "path": "/mnt/datasets/test.csv",
    "url_query": "q=1", "extra": "value"}
    """
    # Handle Windows paths including disk name in this special case
    if re.match(r'^[a-zA-Z]:[\\/]', urlpath):
        return {'protocol': 'file',
                'path': urlpath}

    parsed_path = urlsplit(urlpath)
    protocol = parsed_path.scheme or 'file'
    if parsed_path.fragment:
        path = "#".join([parsed_path.path, parsed_path.fragment])
    else:
        path = parsed_path.path
    if protocol == 'file':
        # Special case parsing file protocol URL on Windows according to:
        # https://msdn.microsoft.com/en-us/library/jj710207.aspx
        windows_path = re.match(r'^/([a-zA-Z])[:|]([\\/].*)$', path)
        if windows_path:
            path = '%s:%s' % windows_path.groups()

    if protocol in ["http", "https"]:
        # for HTTP, we don't want to parse, as requests will anyway
        return {"protocol": protocol, "path": urlpath}

    options = {
        'protocol': protocol,
        'path': path,
    }

    if parsed_path.netloc:
        # Parse `hostname` from netloc manually because `parsed_path.hostname`
        # lowercases the hostname which is not always desirable (e.g. in S3):
        # https://github.com/dask/dask/issues/1417
        options['host'] = parsed_path.netloc.rsplit('@', 1)[-1].rsplit(':', 1)[0]

        if protocol in ("s3", "gcs", "gs"):
            options["path"] = options['host'] + options["path"]
        else:
            options["host"] = options['host']
        if parsed_path.port:
            options['port'] = parsed_path.port
        if parsed_path.username:
            options['username'] = parsed_path.username
        if parsed_path.password:
            options['password'] = parsed_path.password

    if parsed_path.query:
        options['url_query'] = parsed_path.query
    if parsed_path.fragment:
        options['url_fragment'] = parsed_path.fragment

    if inherit_storage_options:
        update_storage_options(options, inherit_storage_options)

    return options


def update_storage_options(options, inherited=None):
    if not inherited:
        inherited = {}
    collisions = set(options) & set(inherited)
    if collisions:
        collisions = '\n'.join('- %r' % k for k in collisions)
        raise KeyError("Collision between inferred and specified storage "
                       "options:\n%s" % collisions)
    options.update(inherited)


# Compression extensions registered via fsspec.compression.register_compression
compressions = {}


def infer_compression(filename):
    """Infer compression, if available, from filename.

    Infer a named compression type, if registered and available, from filename
    extension. This includes builtin (gz, bz2, zip) compressions, as well as
    optional compressions. See fsspec.compression.register_compression.
    """
    extension = os.path.splitext(filename)[-1].strip('.')
    if extension in compressions:
        return compressions[extension]


def build_name_function(max_int):
    """ Returns a function that receives a single integer
    and returns it as a string padded by enough zero characters
    to align with maximum possible integer

    >>> name_f = build_name_function(57)

    >>> name_f(7)
    '07'
    >>> name_f(31)
    '31'
    >>> build_name_function(1000)(42)
    '0042'
    >>> build_name_function(999)(42)
    '042'
    >>> build_name_function(0)(0)
    '0'
    """
    # handle corner cases max_int is 0 or exact power of 10
    max_int += 1e-8

    pad_length = int(math.ceil(math.log10(max_int)))

    def name_function(i):
        return str(i).zfill(pad_length)

    return name_function


def seek_delimiter(file, delimiter, blocksize):
    r"""Seek current file to file start, file end, or byte after delimiter seq.

    Seeks file to next chunk delimiter, where chunks are defined on file start,
    a delimiting sequence, and file end. Use file.tell() to see location afterwards.
    Note that file start is a valid split, so must be at offset > 0 to seek for
    delimiter.

    Parameters
    ----------
    file: a file
    delimiter: bytes
        a delimiter like ``b'\n'`` or message sentinel, matching file .read() type
    blocksize: int
        Number of bytes to read from the file at once.


    Returns
    -------
    Returns True if a delimiter was found, False if at file start or end.

    """

    if file.tell() == 0:
        # beginning-of-file, return without seek
        return False

    # Interface is for binary IO, with delimiter as bytes, but initialize last
    # with result of file.read to preserve compatibility with text IO.
    last = None
    while True:
        current = file.read(blocksize)
        if not current:
            # end-of-file without delimiter
            return False
        full = last + current if last else current
        try:
            if delimiter in full:
                i = full.index(delimiter)
                file.seek(file.tell() - (len(full) - i) + len(delimiter))
                return True
            elif len(current) < blocksize:
                # end-of-file without delimiter
                return False
        except (OSError, ValueError):
            pass
        last = full[-len(delimiter):]


def read_block(f, offset, length, delimiter=None, split_before=False):
    """ Read a block of bytes from a file

    Parameters
    ----------
    f: File
        Open file
    offset: int
        Byte offset to start read
    length: int
        Number of bytes to read, read through end of file if None
    delimiter: bytes (optional)
        Ensure reading starts and stops at delimiter bytestring
    split_before: bool (optional)
        Start/stop read *before* delimiter bytestring.


    If using the ``delimiter=`` keyword argument we ensure that the read
    starts and stops at delimiter boundaries that follow the locations
    ``offset`` and ``offset + length``.  If ``offset`` is zero then we
    start at zero, regardless of delimiter.  The bytestring returned WILL
    include the terminating delimiter string.

    Examples
    --------

    >>> from io import BytesIO  # doctest: +SKIP
    >>> f = BytesIO(b'Alice, 100\\nBob, 200\\nCharlie, 300')  # doctest: +SKIP
    >>> read_block(f, 0, 13)  # doctest: +SKIP
    b'Alice, 100\\nBo'

    >>> read_block(f, 0, 13, delimiter=b'\\n')  # doctest: +SKIP
    b'Alice, 100\\nBob, 200\\n'

    >>> read_block(f, 10, 10, delimiter=b'\\n')  # doctest: +SKIP
    b'Bob, 200\\nCharlie, 300'
    """
    if delimiter:
        f.seek(offset)
        found_start_delim = seek_delimiter(f, delimiter, 2 ** 16)
        if length is None:
            return f.read()
        start = f.tell()
        length -= start - offset

        f.seek(start + length)
        found_end_delim = seek_delimiter(f, delimiter, 2 ** 16)
        end = f.tell()

        # Adjust split location to before delimiter iff seek found the
        # delimiter sequence, not start or end of file.
        if found_start_delim and split_before:
            start -= len(delimiter)

        if found_end_delim and split_before:
            end -= len(delimiter)

        offset = start
        length = end - start

    f.seek(offset)
    b = f.read(length)
    return b


def tokenize(*args, **kwargs):
    """ Deterministic token

    (modified from dask.base)

    >>> tokenize([1, 2, '3'])
    '9d71491b50023b06fc76928e6eddb952'

    >>> tokenize('Hello') == tokenize('Hello')
    True
    """
    if kwargs:
        args += (kwargs,)
    return md5(str(args).encode()).hexdigest()


def stringify_path(filepath):
    """ Attempt to convert a path-like object to a string.

    Parameters
    ----------
    filepath: object to be converted

    Returns
    -------
    filepath_str: maybe a string version of the object

    Notes
    -----
    Objects supporting the fspath protocol (Python 3.6+) are coerced
    according to its __fspath__ method.

    For backwards compatibility with older Python version, pathlib.Path
    objects are specially coerced.

    Any other object is passed through unchanged, which includes bytes,
    strings, buffers, or anything else that's not even path-like.
    """
    if hasattr(filepath, "__fspath__"):
        return filepath.__fspath__()
    elif isinstance(filepath, pathlib.Path):
        return str(filepath)
    return filepath




from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .spec import AbstractFileSystem
from .registry import get_filesystem_class, registry, filesystem
from .mapping import FSMap, get_mapper
from .core import open_files, get_fs_token_paths, open



import pytest
from fsspec.registry import get_filesystem_class, registry


@pytest.mark.parametrize("protocol,module,minversion,oldversion", [
    ("s3", "s3fs", "0.3.0", "0.1.0"),
    ("gs", "gcsfs", "0.3.0", "0.1.0"),
])
def test_minversion_s3fs(protocol, module, minversion, oldversion,
                         monkeypatch):
    registry.clear()
    mod = pytest.importorskip(module, minversion)

    assert get_filesystem_class("s3") is not None
    registry.clear()

    monkeypatch.setattr(mod, "__version__", oldversion)
    with pytest.raises(RuntimeError, match=minversion):
        get_filesystem_class(protocol)



#!/usr/bin/env python
import os

from setuptools import setup
import versioneer

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='fsspec',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
      ],
      description='File-system specification',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/intake/filesystem_spec',
      maintainer='Martin Durant',
      maintainer_email='mdurant@anaconda.com',
      license='BSD',
      keywords='file',
      packages=['fsspec', 'fsspec.implementations'],
      python_requires='>=3.5',
      install_requires=[open('requirements.txt').read().strip().split('\n')],
      zip_safe=False)



import os
import shutil
import pickle
import pytest

import fsspec
from fsspec.implementations.cached import CachingFileSystem
from .test_ftp import ftp_writable, FTPFileSystem


@pytest.fixture
def local_filecache():
    import tempfile
    original_location = tempfile.mkdtemp()
    cache_location = tempfile.mkdtemp()
    original_file = os.path.join(original_location, 'afile')
    data = b'test data'
    with open(original_file, 'wb') as f:
        f.write(data)

    # we can access the file and read it
    fs = fsspec.filesystem('filecache', target_protocol='file',
                           cache_storage=cache_location)

    return (data, original_file, cache_location, fs)


def test_idempotent():
    fs = CachingFileSystem('file')
    fs2 = CachingFileSystem('file')
    assert fs2 is fs
    fs3 = pickle.loads(pickle.dumps(fs))
    assert fs3.storage == fs.storage


def test_worflow(ftp_writable):
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host, port, user, pw)
    with fs.open('/out', 'wb') as f:
        f.write(b'test')
    fs = fsspec.filesystem('cached', target_protocol='ftp',
                           target_options={'host': host, 'port': port,
                                            'username': user, 'password': pw})
    assert os.listdir(fs.storage[-1]) == []
    with fs.open('/out') as f:
        assert os.listdir(fs.storage[-1])
        assert f.read() == b'test'
        assert fs.cached_files[-1]['ftp:///out']['blocks']
    assert fs.cat('/out') == b'test'
    assert fs.cached_files[-1]['ftp:///out']['blocks'] is True

    with fs.open('/out', 'wb') as f:
        f.write(b'changed')

    assert fs.cat('/out') == b'test'  # old value


def test_blocksize(ftp_writable):
    host, port, user, pw = ftp_writable
    fs = FTPFileSystem(host, port, user, pw)
    with fs.open('/out', 'wb') as f:
        f.write(b'test')

    fs = fsspec.filesystem('cached', target_protocol='ftp',
                           target_options={'host': host, 'port': port,
                                            'username': user, 'password': pw})

    assert fs.cat('/out') == b'test'
    with pytest.raises(ValueError):
        fs.open('/out', block_size=1)


def test_local_filecache_creates_dir_if_needed():
    import tempfile
    original_location = tempfile.mkdtemp()
    cache_location = 'foofoobarbar'
    assert not os.path.exists(cache_location)

    try:
        original_file = os.path.join(original_location, 'afile')
        data = b'test data'
        with open(original_file, 'wb') as f:
            f.write(data)

        # we can access the file and read it
        fs = fsspec.filesystem('filecache', target_protocol='file',
                               cache_storage=cache_location)

        with fs.open(original_file, 'rb') as f:
            data_in_cache = f.read()

        assert os.path.exists(cache_location)

    finally:
        shutil.rmtree(cache_location)

    assert data_in_cache == data


def test_local_filecache_basic(local_filecache):
    data, original_file, cache_location, fs = local_filecache

    # reading from the file contains the right data
    with fs.open(original_file, 'rb') as f:
        assert f.read() == data
    assert 'cache' in os.listdir(cache_location)

    # the file in the location contains the right data
    fn = list(fs.cached_files[-1].values())[0]['fn']  # this is a hash value
    assert fn in os.listdir(cache_location)
    with open(os.path.join(cache_location, fn), 'rb') as f:
        assert f.read() == data

    # still there when original file is removed (check=False)
    os.remove(original_file)
    with fs.open(original_file, 'rb') as f:
        assert f.read() == data


def test_local_filecache_does_not_change_when_original_data_changed(local_filecache):
    old_data, original_file, cache_location, fs = local_filecache
    new_data = b'abc'

    with fs.open(original_file, 'rb') as f:
        assert f.read() == old_data

    with open(original_file, 'wb') as f:
        f.write(new_data)

    with fs.open(original_file, 'rb') as f:
        assert f.read() == old_data


def test_local_filecache_gets_from_original_if_cache_deleted(local_filecache):
    old_data, original_file, cache_location, fs = local_filecache
    new_data = b'abc'

    with fs.open(original_file, 'rb') as f:
        assert f.read() == old_data

    with open(original_file, 'wb') as f:
        f.write(new_data)

    shutil.rmtree(cache_location)
    assert os.path.exists(original_file)

    with open(original_file, 'rb') as f:
        assert f.read() == new_data

    with fs.open(original_file, 'rb') as f:
        assert f.read() == new_data

    # the file in the location contains the right data
    fn = list(fs.cached_files[-1].values())[0]['fn']  # this is a hash value
    assert fn in os.listdir(cache_location)
    with open(os.path.join(cache_location, fn), 'rb') as f:
        assert f.read() == new_data


def test_local_filecache_with_new_cache_location_makes_a_new_copy(local_filecache):
    import tempfile
    data, original_file, old_cache_location, old_fs = local_filecache
    new_cache_location = tempfile.mkdtemp()

    with old_fs.open(original_file, 'rb') as f:
        assert f.read() == data

    new_fs = fsspec.filesystem('filecache', target_protocol='file',
                               cache_storage=new_cache_location)

    with new_fs.open(original_file, 'rb') as f:
        assert f.read() == data

    # the file in the location contains the right data
    fn = list(new_fs.cached_files[-1].values())[0]['fn']  # this is a hash value
    assert fn in os.listdir(old_cache_location)
    assert fn in os.listdir(new_cache_location)

    with open(os.path.join(new_cache_location, fn), 'rb') as f:
        assert f.read() == data


def test_filecache_multicache():
    import tempfile
    origin = tempfile.mkdtemp()
    cache1 = tempfile.mkdtemp()
    cache2 = tempfile.mkdtemp()
    data = b'test data'
    f1 = os.path.join(origin, 'afile')
    f2 = os.path.join(origin, 'bfile')
    with open(f1, 'wb') as f:
        f.write(data)
    with open(f2, 'wb') as f:
        f.write(data * 2)

    # populates first cache
    fs = fsspec.filesystem('filecache', target_protocol='file',
                           cache_storage=cache1)
    assert fs.cat(f1) == data

    assert len(os.listdir(cache1)) == 2  # cache and hashed afile
    assert len(os.listdir(cache2)) == 0  # hasn't been intialized yet

    # populates last cache if file not found in first cache
    fs = fsspec.filesystem('filecache', target_protocol='file',
                           cache_storage=[cache1, cache2])

    assert fs.cat(f1) == data
    assert fs.cat(f2) == data * 2

    assert 'cache' in os.listdir(cache1)
    assert 'cache' in os.listdir(cache2)

    cache1_contents = [f for f in os.listdir(cache1) if f != 'cache']
    assert len(cache1_contents) == 1

    with open(os.path.join(cache1, cache1_contents[0]), 'rb') as f:
        assert f.read() == data

    cache2_contents = [f for f in os.listdir(cache2) if f != 'cache']
    assert len(cache2_contents) == 1

    with open(os.path.join(cache2, cache2_contents[0]), 'rb') as f:
        assert f.read() == data * 2


def test_filecache_multicache_with_same_file_different_data_reads_from_first():
    import tempfile
    origin = tempfile.mkdtemp()
    cache1 = tempfile.mkdtemp()
    cache2 = tempfile.mkdtemp()
    data = b'test data'
    f1 = os.path.join(origin, 'afile')
    with open(f1, 'wb') as f:
        f.write(data)

    # populate first cache
    fs = fsspec.filesystem('filecache', target_protocol='file',
                           cache_storage=cache1)
    assert fs.cat(f1) == data

    with open(f1, 'wb') as f:
        f.write(data * 2)

    # populate second cache
    fs = fsspec.filesystem('filecache', target_protocol='file',
                           cache_storage=cache2)

    assert fs.cat(f1) == data * 2

    # the filenames in each cache are the same, but the data is different
    assert os.listdir(cache1) == os.listdir(cache2)

    fs = fsspec.filesystem('filecache', target_protocol='file',
                           cache_storage=[cache1, cache2])

    assert fs.cat(f1) == data


def test_filecache_with_checks():
    import tempfile
    import time
    origin = tempfile.mkdtemp()
    cache1 = tempfile.mkdtemp()
    data = b'test data'
    f1 = os.path.join(origin, 'afile')
    with open(f1, 'wb') as f:
        f.write(data)

    # populate first cache
    fs = fsspec.filesystem('filecache', target_protocol='file',
                           cache_storage=cache1, expiry_time=0.1)
    fs2 = fsspec.filesystem('filecache', target_protocol='file',
                            cache_storage=cache1, check_files=True)
    assert fs.cat(f1) == data
    assert fs2.cat(f1) == data

    with open(f1, 'wb') as f:
        f.write(data * 2)

    assert fs.cat(f1) == data  # does not change
    assert fs2.cat(f1) == data * 2  # changed, since origin changed
    time.sleep(0.11)  # allow cache details to expire
    assert fs.cat(f1) == data * 2  # changed, since origin changed
